import os
from openProduction.common import Signals

class TestParams:
    DEVICE_ID = "deviceID"
    def __init__(self, params, folder):
        self._params = params
        self._paramFolder = folder
        self._files = []
        self.keyChanged = Signals.Signal()
        self.deviceIDSet = Signals.Signal()
        self._readOnly = False
        
    def setReadOnly(self, ro):
        self._readOnly = ro
        
    def removeKey(self, key):
        if self._readOnly:
            raise RuntimeError("remove key %s not allowed (TestParams is read-only)"%key)
        self._params.pop(key, None)
        
    def __getitem__(self, key):       
        return self._params[key]
    
    def __setitem__(self, key, value):
        if self._readOnly:
            raise RuntimeError("setting %s not allowed (TestParams is read-only)"%key)
            
        hasChanged = False
        if key not in self._params:
            hasChanged = True
        else:
            if value != self._params[key]:
                hasChanged = True

        self._params[key] = value
        
        if hasChanged:
            self.keyChanged.emit(key, value)
        if key == self.DEVICE_ID:
            self.deviceIDSet.emit(value)
        
    def __contains__(self, key):
        return key in self._params
    
    def move(self, key):
        if self._readOnly:
            raise RuntimeError("move not allowed (TestParams is read-only)")
        params = {}
        if self.DEVICE_ID in self._params:
            devID = self._params[self.DEVICE_ID]
            self._params.pop(self.DEVICE_ID, None)
        else:
            devID = None
        
        params[key] = self._params
        self._params = params
        if devID != None:
            self._params[self.DEVICE_ID] = devID
            

    def setDeviceID(self, val):
        if self._readOnly:
            raise RuntimeError("setting deviceID not allowed (TestParams is read-only)")
        self._params[self.DEVICE_ID] = val
        self.deviceIDSet.emit(val)
        
    def getDeviceID(self):
        return self._params[self.DEVICE_ID]
    
    def clear(self):
        self._params = {}
        self._files = []
    
    def abspath(self, key):
        if self._paramFolder != None:
            return os.path.join(self._paramFolder, self._params[key])
        else:
            return self._params[key]
    
    def addFile(self, fileName):
        if self._readOnly:
            raise RuntimeError("adding file not allowed (TestParams is read-only)")
        fName = os.path.abspath(fileName)
        if fName in self._files:
            raise RuntimeError("file %s already exists in params"%fName)
        if os.path.exists(fName) == False:
            raise RuntimeError("file %s doesn't exist"%fName)
        self._files.append(fName)
    
    def getParams(self):
        return self._params
    
    def getFiles(self):
        return self._files
    
    def __len__(self):
        return len(self._params)
    
    def __str__(self):
        MAX_LEN = 40
        
        maxName = len("Name")
        maxValue = len("Value")
        
        for key, val in self._params.items():
            if type(val) == type({}):
                for subKey, subVal in val.items():
                    subKey = key+"."+subKey
                    maxName = max(maxName, len(subKey))
                    maxValue = max(maxValue, len(str(subVal)))
            else:
                maxName = max(maxName, len(key))
                maxValue = max(maxValue, len(str(val)))
            
        if maxName > MAX_LEN:
            maxName = MAX_LEN
        if maxValue > MAX_LEN:
            maxValue = MAX_LEN
            
        colName = maxName+2
        colVal = maxValue+2
        
        myStr = ""
        myStr += ("+"+"-"*colName + "+" + "-"*colVal + "+") + "\n"
        myStr += ("| Name " + " "*(colName-6) + "| Value " + " "*(colVal-7) + "|") + "\n"
        myStr += ("+"+"="*colName + "+" + "="*colVal + "+") + "\n"
        
        def toStrKeyVal(key,val):
            myStr = ""
            printKey = key
            if len(printKey)>MAX_LEN:
                printKey = printKey[:MAX_LEN-3]+"..."
            printVal = str(val)
            if len(printVal)>MAX_LEN:
                printVal = printVal[:MAX_LEN-3]+"..."
                
            myStr += ("| " + printKey + " "*(colName-len(printKey)-1) + "| " +
                      str(printVal) +" "*(colVal-len(str(printVal))-1) + "|" ) + "\n"
            myStr += ("+"+"-"*colName + "+" + "-"*colVal + "+") + "\n"
            return myStr
        
        for key, val in self._params.items():
            if type(val) == type({}):
                for subKey, subVal in val.items():
                    subKey = key+"."+subKey
                    myStr += toStrKeyVal(subKey, subVal)
            else:
                myStr += toStrKeyVal(key, val)
                
        for f in self._files:
            myStr += toStrKeyVal("**FILE**", f)
                
        return myStr


VERSION_SEPARATOR = "."
TAG_SEPARATOR = "/"
URL_SEPARATOR = "|"

def isValidStr(string):
    if string.count(VERSION_SEPARATOR) == 0 and \
       string.count(TAG_SEPARATOR) == 0 and \
       string.count(URL_SEPARATOR) == 0:
        return True
    else:
        return False

def invalidChars():
    return VERSION_SEPARATOR + TAG_SEPARATOR + URL_SEPARATOR


from openProduction.suite import TestSuite, TestClass
from openProduction.common import misc
import glob
import os
from importlib.machinery import SourceFileLoader
import logging


class TestGroupLoaderClass:
    
    def __init__(self, inst):
        if isinstance(inst, TestClass.TestClass):
            self._tcInst = inst
        else:
            raise RuntimeError("Test class has to be of type "+
                               "openProduction.suite.TestClass.TestClass")
    
    def discover(self, patternRegular="step_", 
                 patternSetup = "setup_", patternTearDown = "teardown_",
                 patternLoad = "load_", patternUnload = "unload_"):
        className = self._tcInst.__class__.__name__
        classDoc = self._tcInst.__class__.__doc__
        
        tcg = TestSuite.TestCaseGroup(className, classDoc, self._tcInst)
        for val in dir(self._tcInst):
            if val.find(patternRegular) == 0:
                typ = TestSuite.StepType.REGULAR
            elif val.find(patternSetup) == 0:
                typ = TestSuite.StepType.SETUP
            elif val.find(patternTearDown) == 0:
                typ = TestSuite.StepType.TEARDOWN
            elif val.find(patternLoad) == 0:
                typ = TestSuite.StepType.LOAD
            elif val.find(patternUnload) == 0:
                typ = TestSuite.StepType.UNLOAD
            else:
                typ = None
                
            if typ != None:
                
                methodName = val
                func = getattr(self._tcInst, val)
                doc = func.__doc__
                hint = TestSuite.InstallHints()
                tc = TestSuite.TestCase(methodName, doc, func, typ, hint)
                tcg.addTestCase(tc)
        return tcg
    
class TestDirectoryLoader:
    
    def __init__(self, directory):
        self.logger = logging.getLogger(misc.getAppName())
        self.dir = directory
        
    def discover(self, ctor, pattern=["hook_*.py"], testClass=TestClass.TestClass,
                 patternRegular="step_", patternSetup = "setup_",
                 patternTearDown = "teardown_", patternLoad = "load_",
                 patternUnload = "unload_", dontConstruct=False):
        filesGrabbed = []
        for files in pattern:
            join = os.path.join(self.dir, files)
            filesGrabbed.extend(glob.glob(join))
        
        tgs = []
        for file in filesGrabbed:
            self.logger.info("discoverying file %s for TestClasses"%file)
            loader = TestClassLoader(file)
            tgs.extend(loader.discover(ctor, testClass, patternRegular, patternSetup,
                 patternTearDown, patternLoad, patternUnload, dontConstruct))
            
        return tgs
    
class TestClassLoader:
    
    def __init__(self, fname):
        self.fname = fname
        
    def discover(self, ctor, testClass=TestClass.TestClass,
                 patternRegular="step_", patternSetup = "setup_",
                 patternTearDown = "teardown_", patternLoad = "load_",
                 patternUnload = "unload_", dontConstruct=False,
                 reservedNames=["TestClass", "ProductExecutor"]):
        testClasses = []
        f = os.path.splitext(os.path.split(self.fname)[-1])[0]
        mod = SourceFileLoader("TestClassLoader_%s"%f, self.fname).load_module()
        classes = dir(mod)
        for cla in classes:
            if cla in reservedNames:
                continue
            claObj = getattr(mod, cla)
            try:
                if issubclass(claObj, testClass):
                    testClasses.append(claObj)
            except:
                pass
            
        if dontConstruct:
            return testClasses
        
        tgs = []
        objs = []
        for cla in testClasses:
            objs.append(cla(ctor))
            
        for obj in objs:
            loader = TestGroupLoaderClass(obj)
            tgs.append(loader.discover(patternRegular, patternSetup, patternTearDown, patternLoad, patternUnload))
            
        return tgs
            
class TestModuleLoader:
    
    def __init__(self, module):
        self.module = module
        
    def discover(self, ctor, testClass=TestClass.TestClass,
                 patternRegular="step_", patternSetup = "setup_",
                 patternTearDown = "teardown_", patternLoad = "load_",
                 patternUnload = "unload_", dontConstruct=False):
        testClasses = []

        mod = self.module
        classes = dir(mod)
        for cla in classes:
            claObj = getattr(mod, cla)
            try:
                if issubclass(claObj, testClass):
                    testClasses.append(claObj)
            except:
                pass
            
        if dontConstruct:
            return testClasses
        
        tgs = []
        objs = []
        for cla in testClasses:
            objs.append(cla(ctor))
            
        for obj in objs:
            loader = TestGroupLoaderClass(obj)
            tgs.append(loader.discover(patternRegular, patternSetup, patternTearDown, patternLoad, patternUnload))
            
        return tgs    



import sys
from openProduction.suite.TestSuite import TestSuiteRunner, ExecutionResult, TestSuiteIO, SuiteState
from openProduction.common import misc
import colorama
colorama.init()

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

class IOHandlerCli(TestSuiteIO.BaseIOHandler):
    def __init__(self):
        super(IOHandlerCli, self).__init__()
    
    def _queryYesNo(self, title, msg, timeout=10):
        rv = misc.queryYesNo(msg)
        if rv:
            return TestSuiteIO.IOAnswer.YES
        else:
            return TestSuiteIO.IOAnswer.NO
        
    def message(self, msg):
        print(msg)

class TestSuiteCliRunner:
    
    def __init__(self, testSuite=None, testSuiteRunner=None, playSound=True):
        if testSuiteRunner == None:
            self._testSuiteRunner = TestSuiteRunner(testSuite)                
        else:
            self._testSuiteRunner = testSuiteRunner
        self._testSuiteRunner.sequenceStart.connect(self._onSequenceStart)
        self._testSuiteRunner.sequenceComplete.connect(self._onSequenceComplete)
        self._testSuiteRunner.stepStart.connect(self._onStepStart)
        self._testSuiteRunner.stepComplete.connect(self._onStepComplete)
        self._playSound = playSound
        
    def run(self):
        fut = self._testSuiteRunner.start()
        def _onReady(future):
            if future.exception() != None:
                print("Something went wrong")
                print(future.exception())
        fut.add_done_callback(_onReady)
                
    def _onSequenceComplete(self, state, suiteResult):
        self._reportErrors(suiteResult)
        self._reportOverall(suiteResult)
        
        attrs = suiteResult.suite.values
        if len(attrs) > 0:
            print("")
            print("----------------------------------------------------------------------")
            print("The test suite contains the following values:")
            print(attrs)
            print("----------------------------------------------------------------------")
        
        print("======================================================================")
        print("= STAGE %s complete"%(state.name) + " "*(52-len(state.name)) + "=")
        print("======================================================================")
        sys.stdout.flush()
        
        if self._playSound:
            if state == SuiteState.REGULAR:
                ioHandler = TestSuiteIO.BaseIOHandler()
                if suiteResult.isSuccessful() == False:
                    ioHandler.playError()
                else:
                    ioHandler.playSuccess()
        
    def _reportErrors(self, suiteResult):
        for res in suiteResult.stepResults:
            if res.result == ExecutionResult.EXCEPTION:
                failStr = "ERROR"
            elif res.result == ExecutionResult.FAIL:
                failStr = "FAIL"
            elif res.result == ExecutionResult.ABORT:
                failStr = "ABORT"
            else:
                continue
            print("======================================================================")
            print("%s: %s (%s)\n%s"%(failStr, res.case.name, res.case.group.name, res.case.doc))
            print("----------------------------------------------------------------------")
            print(res.msg)
        
    def _reportOverall(self, suiteResult, preStr=""):
        print("----------------------------------------------------------------------")
        dur = suiteResult.executionTime
        numTests = suiteResult.numTotal
        numGood = suiteResult.numGood
        numFail = suiteResult.numFails
        numError = suiteResult.numErrors
        numSkip = suiteResult.numSkips
        print("Ran %d step(s) in %.03f s"%(numTests, dur))
        print("")
        if suiteResult.isSuccessful() == False:
            
            overall = preStr + bcolors.FAIL+"FAILED"+bcolors.ENDC+" (failures=%d, errors=%d, skip=%d, ok=%d)"%(numFail,
                              numError, numSkip, numGood)
        else:
            overall = preStr + bcolors.OKGREEN+"OK"+bcolors.ENDC+" (failures=%d, errors=%d, skip=%d, ok=%d)"%(numFail,
                              numError, numSkip, numGood)
        print(overall)
    
    def _onSequenceStart(self, state):
        state = str(state)
        print("======================================================================")
        print("= Starting STAGE %s"%(state) + " "*(52-len(state)) + "=")
        print("======================================================================")
    
    def _onStepStart(self, state, testCase, idx, numCases):
        sys.stdout.write(testCase.getFullName() + "(%s): ... "%testCase.doc)
        sys.stdout.flush()
    
    def _onStepComplete(self, state, stepRes, idx, numCases):
        if stepRes.result == ExecutionResult.EXCEPTION:
            resStr = "Exception"
        elif stepRes.result == ExecutionResult.FAIL:
            resStr = "Failure"
        elif stepRes.result == ExecutionResult.SUCCESS:
            resStr = "Ok"
        elif stepRes.result == ExecutionResult.SKIP:
            resStr = "Skip"
        elif stepRes.result == ExecutionResult.ABORT:
            resStr = "Abort"
        else:
            resStr = str(stepRes.result)
            
        sys.stdout.write(resStr + "\n")
        sys.stdout.flush()



#!/usr/bin/env python
#
# Short & sweet script for use with git clone and fetch credentials.
# Requires GIT_USERNAME and GIT_PASSWORD environment variables,
# intended to be called by Git via GIT_ASKPASS.
#

from sys import argv
from os import environ

if "USERNAME" in argv[1].upper():
    print (environ['GIT_USERNAME'])
    exit()

if "PASSWORD" in argv[1].upper():
    print (environ['GIT_PASSWORD'])
    exit()

exit(1)


import openProduction.connectors.BaseConnector as plugintypes
from openProduction.common import Types, misc
import socket
import logging
import traceback
import pymysql
import json

class MySQLConnector(plugintypes.BaseConnector):    
    def __init__(self):
        __fields__ = [Types.String("user", "Username to access SQL database", ""),
                      Types.Password("password", "Password to access SQL database (WARNING: do not use a root-db password. This password is stored in plain-text)", ""),
                      Types.String("hostname", "Hostname (or IP) of SQL database", ""),
                      Types.Int("port", "Port of SQL database", "", minVal=1, maxVal=65535)
                      ]
        super(MySQLConnector, self).__init__(__fields__)
        self._unconfigured = True
        self.logger = logging.getLogger(misc.getAppName())
        self.err = ""
        self.transactionOpen = False
        
    def getValueDataType(self, value):
        dt = ""
        if isinstance(value, Types.Int):
            dt = "INT"
        elif isinstance(value, Types.Bool):
            dt = "BIT"
        elif isinstance(value, Types.Float):
            dt = "FLOAT"
        elif isinstance(value, Types.String):
            dt = "TEXT"
        elif isinstance(value, Types.Date):
            dt = "DATETIME"
        else:
            raise RuntimeError("unsupported data type %s"%Types.String)
        return dt
    
    def getDescription(self):
        return "This connector uses your local (or a shared) file system"
    
    def listDatabases(self):
        connection = pymysql.connect(host=self.host, user=self.username, port=self.port, password=self.password, charset=self.charset, cursorclass=pymysql.cursors.DictCursor)
        sql = "SHOW DATABASES;"        
        c = connection.cursor()
        r = c.execute(sql)
        if r > 0:
            rv = c.fetchall()
        else:
            rv = []
        
        return rv
    
    def deleteRows(self, dbname, table, filt):
        sql = "DELETE FROM `%s`.`%s` WHERE %s;"%(dbname, table, filt)
        rv, data = self._execSQLWithFetch(sql)
        if rv == plugintypes.ConnectorErrors.NO_ERROR and data == None:
            rv = plugintypes.ConnectorErrors.NOT_FOUND
        return rv
        
    def deleteDatabase(self, dbName):
        sql = "DROP DATABASE %s;"%dbName
        rv, rowID = self._execSQLWithoutFetch(sql)
        return rv
    
    def createDatabase(self, dbName):
        notExists = ""    
        sql = "CREATE DATABASE %s %s;"%(notExists, dbName)
        rv, rowID = self._execSQLWithoutFetch(sql)
        return rv
    
    def createTables(self, dbName, tables):
        sqls = []
        params = []
        for tab in tables:
            sql, param = self._createTable(dbName, tab)
            sqls.append(sql)
            params.append(param)
        
        rv, rowID = self._execSQLWithoutFetch(sqls, params=params)
        return rv
    
    def createTable(self, dbName, tableName, columns):
        tables = [{"columns": columns, "name": tableName}]
        return self.createTables(dbName, tables)
    
    def _createTable(self, dbName, table):
        primaryKey = []
        references = []
        lines = []
        columns = table["columns"]
        tableName = table["name"]
        escapes = []
        jsonChecks = []
        for col in columns:
            colName = col["name"]
            dtype = col["type"]
            notNull = ""
            if "notNull" in col:
                if col["notNull"] == True:
                    notNull = "NOT NULL"
            autoInc = ""
            if "autoIncrement" in col:
                if col["autoIncrement"] == True:
                    autoInc = "AUTO_INCREMENT"
            unique = ""
            if "unique" in col:
                if col["unique"] == True:
                    unique = "UNIQUE"
            default = ""
            if "default" in col:
                default = "DEFAULT %s"
                escapes.append(str(col["default"]))
            if "constraint" in col:
                constraint = col["constraint"]
            else:
                constraint = ""
            line = "`%s` %s %s %s %s %s %s"%(colName, dtype, notNull, unique, autoInc, default, constraint)
            lines.append(line)
            
            if "primaryKey" in col:
                if col["primaryKey"] == True:
                    primaryKey.append(colName)
                    
            if "foreign" in col:
                ref = "FOREIGN KEY (`%s`) REFERENCES `%s`(`%s`) ON DELETE CASCADE ON UPDATE CASCADE"%(colName, col["foreign"][0],col["foreign"][1])
                references.append(ref)
                
            if dtype.lower() == "json":
                jsonChecks.append("CHECK (JSON_VALID(%s))"%colName)
        
        if primaryKey != []:
            keys = ",".join(primaryKey)
            primaryKey = "PRIMARY KEY (%s)"%keys
            lines.append(primaryKey)
            
        uniqueLines = []
        if "uniqueKey" in table:
            if type(table["uniqueKey"]) == type([]):
                #list of lists?
                if type(table["uniqueKey"][0]) == type([]):        
                    for uniques in table["uniqueKey"]:
                        uniqueCols = ['`' +x+'`' for x in uniques]
                        uniqueLines.append("UNIQUE KEY(%s)"%(",".join(uniqueCols)))
                else:
                    uniqueCols = ['`' +x+'`' for x in table["uniqueKey"]]
                    uniqueLines.append("UNIQUE KEY(%s)"%(",".join(uniqueCols)))

        lines = lines + references + uniqueLines + jsonChecks

        content = ",\n".join(lines)
        sql = "CREATE TABLE `%s`.`%s` (%s);"%(dbName, tableName, content)
        # print(sql)
        return sql, escapes
    
    def changeColumnDataType(self, dbName, tableName, columnName, dataTypeString):
        sql = "ALTER TABLE %s.%s MODIFY COLUMN %s %s;"%(dbName, tableName, columnName, dataTypeString)
        rv, rowID = self._execSQLWithoutFetch(sql)
        return rv
    
    def createColumn(self, dbName, tableName, columnName, dataTypeString, constraints=None):
        if constraints == None:
            constraints = ""
        sql = "ALTER TABLE %s.%s ADD %s %s %s;"%(dbName, tableName, columnName, dataTypeString, constraints)
        rv, rowID = self._execSQLWithoutFetch(sql)
        return rv
    
    def saveValues(self, dbName, table, values):
        colNames = []
        valNames = []
        vals = []
        for val in values:
            colNames.append("`" + val.name + "`")
            valNames.append("%s")
            vals.append(val.value)
        
        valNames = (", ").join(valNames)
        colNames = (", ").join(colNames)
        sql = "INSERT INTO `%s`.`%s` (%s) VALUES (%s);"%(dbName, table, colNames, valNames)
        return self._execSQLWithoutFetch(sql, params=vals)
    
    def addRow(self, dbName, table, entries):        
        colNames = []
        valArr = []
        vals = []
        
        if type(entries) == type([]):
            for val in entries:
                colNames.append("`" + val.name + "`")
                valArr.append("%s")
                vals.append(val.value)
        elif type(entries) == type({}):
            for key, val in entries.items():
                colNames.append("`" + key + "`")
                valArr.append("%s")
                if type(val) == type({}):
                    val = json.dumps(val)
                vals.append(val)
        
        valArr = (", ").join(valArr)
        colNames = (", ").join(colNames)
        sql = "INSERT INTO `%s`.`%s` (%s) VALUES (%s);"%(dbName, table, colNames, valArr)
        return self._execSQLWithoutFetch(sql, params=vals)
    
    def updateRow(self, dbName, table, entries, filt):
        colNames = []
        vals = []
        
        if type(entries) == type([]):
            for val in entries:
                colNames.append(val.name)
                vals.append(val.value)
        elif type(entries) == type({}):
            for key, val in entries.items():
                colNames.append(key)
                if type(val) == type({}):
                    val = json.dumps(val)
                vals.append(val)
        
        cols = ["%s"%x+"=%s" for x in colNames]
        values = ",".join(cols)
        sql = "UPDATE `%s`.`%s` SET %s WHERE %s;"%(dbName, table, values, filt)
        rv, row_id = self._execSQLWithoutFetch(sql, params=vals)
        return rv
        
    def getData(self, dbName, table, columns, filt=None, joins=None, orderBy=None):
        colNames = (", ").join(columns)
        if filt != None:
            filt = "WHERE " + filt
        else:
            filt = ""
        if joins != None:
             joins = ["INNER JOIN "+x for x in joins]
             joins = "\n".join(joins)
        else:
            joins = ""
            
        if orderBy != None:
            order = "ORDER BY %s"%orderBy
        else:
            order = ""
            
        sql = "SELECT %s from `%s`.`%s` %s %s %s;"%(colNames, dbName, table, joins, filt, order)
        return self._execSQLWithFetch(sql)
    
    def getColumnDataType(self, dbName, table, columnName):
        sql = "SELECT `DATA_TYPE` \
FROM `INFORMATION_SCHEMA`.`COLUMNS` \
WHERE `TABLE_SCHEMA`='%s' \
AND `TABLE_NAME`='%s' \
AND `COLUMN_NAME`='%s';"%(dbName, table, columnName) 
        ok, rv = self._execSQLWithFetch(sql)
        
        colType = ""
        if ok == plugintypes.ConnectorErrors.NO_ERROR and rv != None:
            if len(rv) > 0:
                colType = rv[0]["DATA_TYPE"].upper()
        return ok, colType
    
    def dataBaseExists(self, dbName):
        ok = True
        sql = "SELECT SCHEMA_NAME \
FROM INFORMATION_SCHEMA.SCHEMATA \
WHERE SCHEMA_NAME = '%s';"%dbName
        ok, rv = self._execSQLWithFetch(sql)
        if ok == plugintypes.ConnectorErrors.NO_ERROR and rv != None:
            ok = False
        return ok
    
    def tableExists(self, dbName, table):
        ok = True
        sql = "SELECT *  \
FROM information_schema.tables \
WHERE table_schema = '%s'  \
AND table_name = '%s'"%(dbName, table)
        ok, rv = self._execSQLWithFetch(sql)
        if ok == plugintypes.ConnectorErrors.NO_ERROR and rv != None:
            ok = False
        return ok
    
    def getColumnNames(self, dbName, table):
        colNames = []
        
        sql = "SELECT `COLUMN_NAME` \
FROM `INFORMATION_SCHEMA`.`COLUMNS` \
WHERE `TABLE_SCHEMA`='%s' \
AND `TABLE_NAME`='%s';"%(dbName, table)
        ok, rv = self._execSQLWithFetch(sql)
        
        if ok == plugintypes.ConnectorErrors.NO_ERROR and rv != None:
            for col in rv:
                colNames.append(col["COLUMN_NAME"])

        return ok, colNames
    
    def _execSQLWithFetch(self, sql, db=None, params=[]):
        data = None
        rv = plugintypes.ConnectorErrors.NO_ERROR
        
        con = pymysql.connect(host=self.host,
                             user=self.username,
                             port=self.port,
                             password=self.password,
                             charset=self.charset,
                             db=db,
                             max_allowed_packet=32*1024*1024,
                             cursorclass=pymysql.cursors.DictCursor)               
        
        try:
            c = con.cursor()
            r = c.execute(sql, params)
            con.commit()
            if r > 0:
                data =  c.fetchall()
        except pymysql.err.OperationalError as e:
            if e.args[0] == 1044:
                rv = plugintypes.ConnectorErrors.FORBIDDEN
            else:
                self.logger.error("SQL statement '%s' failed with error message %s"%(sql, str(e)))
                self.logger.info("full traceback:\n%s"%traceback.format_exc())
                rv = plugintypes.ConnectorErrors.UNSPECIFIC
        except pymysql.err.ProgrammingError as e:
            if e.args[0] == 1007:
                rv = plugintypes.ConnectorErrors.EXISTS
            else:
                self.logger.error("SQL statement '%s' failed with error message %s"%(sql, str(e)))
                self.logger.info("full traceback:\n%s"%traceback.format_exc())
                rv = plugintypes.ConnectorErrors.UNSPECIFIC
        except pymysql.err.IntegrityError as e:
            if e.args[0] == 1062:
                rv = plugintypes.ConnectorErrors.EXISTS
            else:
                self.logger.error("SQL statement '%s' failed with error message %s"%(sql, str(e)))
                self.logger.info("full traceback:\n%s"%traceback.format_exc())
                rv = plugintypes.ConnectorErrors.UNSPECIFIC
        except pymysql.err.InternalError as e:
            if e.args[0] == 1008:
                rv = plugintypes.ConnectorErrors.NOT_FOUND
            else:
                self.logger.error("SQL statement '%s' failed with InternalError error message %s"%(sql, str(e)))
                self.logger.info("full traceback:\n%s"%traceback.format_exc())
                rv = plugintypes.ConnectorErrors.UNSPECIFIC                
        except Exception as e:
            rv = plugintypes.ConnectorErrors.UNSPECIFIC
            self.logger.error("SQL statement '%s' failed with error message %s"%(sql, str(e)))
            self.logger.info("full traceback:\n%s"%traceback.format_exc())     
        finally:
            c.close()
            
        return rv, data
    
    def startTransaction(self, db):
        rv = plugintypes.ConnectorErrors.NO_ERROR
        try:
            self.transactionOpen = True
            self.connection = pymysql.connect(host=self.host,
                                 user=self.username,
                                 port=self.port,
                                 password=self.password,
                                 charset=self.charset,
                                 db=db,
                                 max_allowed_packet=32*1024*1024,
                                 cursorclass=pymysql.cursors.DictCursor)
            self.connection.autocommit = False
        except Exception as e:
            rv = plugintypes.ConnectorErrors.UNSPECIFIC
            self.logger.error("mysql connection failed with %s error message %s"%(type(e), str(e)))
            self.logger.info("full traceback:\n%s"%traceback.format_exc())
        return rv
        
    def commitTransaction(self):
        rv = plugintypes.ConnectorErrors.NO_ERROR
        try:
            self.connection.commit()
        except Exception as e:
            rv = plugintypes.ConnectorErrors.UNSPECIFIC
            self.logger.error("mysql commit failed with %s error message %s"%(type(e), str(e)))
            self.logger.info("full traceback:\n%s"%traceback.format_exc())
            self.connection.rollback()
            self.logger.info("sql transaction rollback due to previsous commit error")
        finally:
            self.connection.close()
            self.transactionOpen = False
        return rv
    
    def rollbackTransaction(self):
        rv = plugintypes.ConnectorErrors.NO_ERROR
        try:
            self.connection.rollback()
        except Exception as e:
            rv = plugintypes.ConnectorErrors.UNSPECIFIC
            self.logger.error("mysql commit failed with %s error message %s"%(type(e), str(e)))
            self.logger.info("full traceback:\n%s"%traceback.format_exc())
        finally:
            self.connection.close()
            self.transactionOpen = False
        return rv
        
    def _execSQLWithoutFetch(self, sql, db=None, params=[]):
        rowID = None
        rv = plugintypes.ConnectorErrors.NO_ERROR
        
        if self.transactionOpen == False:
            connection = pymysql.connect(host=self.host,
                                 user=self.username,
                                 port=self.port,
                                 password=self.password,
                                 charset=self.charset,
                                 db=db,
                                 max_allowed_packet=32*1024*1024,
                                 cursorclass=pymysql.cursors.DictCursor)
        else:
            connection = self.connection
        
        try:
            with connection.cursor() as cursor:
                # Create a new record
                if type(sql) == type([]):
                    for idx in range(len(sql)):
                        cursor.execute(sql[idx], params[idx])
                else:
                    cursor.execute(sql, params)
        
            rowID = cursor.lastrowid
            
            if self.transactionOpen == False:
                connection.commit()
        except pymysql.err.OperationalError as e:
            if e.args[0] == 1044:
                rv = plugintypes.ConnectorErrors.FORBIDDEN
            else:
                self.logger.error("SQL statement '%s' failed with OperationalError error message %s"%(sql, str(e)))
                self.logger.info("full traceback:\n%s"%traceback.format_exc())
                rv = plugintypes.ConnectorErrors.UNSPECIFIC
        except pymysql.err.ProgrammingError as e:
            if e.args[0] == 1007:
                rv = plugintypes.ConnectorErrors.EXISTS
            else:
                self.logger.error("SQL statement '%s' failed with ProgrammingError error message %s"%(sql, str(e)))
                self.logger.info("full traceback:\n%s"%traceback.format_exc())
                rv = plugintypes.ConnectorErrors.UNSPECIFIC
        except pymysql.err.IntegrityError as e:
            if e.args[0] == 1062:
                rv = plugintypes.ConnectorErrors.EXISTS
            else:
                self.logger.error("SQL statement '%s' failed with IntegrityError error message %s"%(sql, str(e)))
                self.logger.info("full traceback:\n%s"%traceback.format_exc())
                rv = plugintypes.ConnectorErrors.UNSPECIFIC
        except pymysql.err.InternalError as e:
            if e.args[0] == 1008:
                rv = plugintypes.ConnectorErrors.NOT_FOUND
            else:
                self.logger.error("SQL statement '%s' failed with InternalError error message %s"%(sql, str(e)))
                self.logger.info("full traceback:\n%s"%traceback.format_exc())
                rv = plugintypes.ConnectorErrors.UNSPECIFIC
        except Exception as e:
            rv = plugintypes.ConnectorErrors.UNSPECIFIC
            self.logger.error("SQL statement '%s' failed with %s error message %s"%(sql, type(e), str(e)))
            self.logger.info("full traceback:\n%s"%traceback.format_exc())
        finally:
            if self.transactionOpen == False:
                connection.close()
            
        return rv, rowID
    
    def errmsg(self):
        return self.err
    
    def _onNewConfig(self):
        rv = plugintypes.ConnectorErrors.NO_ERROR
        self.port = self.config["port"]
        self.username = self.config["user"]
        self.password = self.config["password"]
        self.charset = 'utf8'
            
        host = self.config["hostname"]
        try:
            self.logger.info("trying to resolve host %s"%host)
            self.host = socket.gethostbyname(host)
        except:
            rv = plugintypes.ConnectorErrors.NOT_FOUND
            self.logger.error("Error resolving host %s"%str(host))
            self.logger.info("full traceback:\n%s"%traceback.format_exc())
            
        if rv == plugintypes.ConnectorErrors.NO_ERROR:
            self.logger.info("host %s resolved to IP %s"%(host, self.host))
            try:
                self.logger.info("trying to connect to db-server %s:%d"%(self.host, self.port))    
                pymysql.connect(host=self.host, user=self.username, port=self.port, password=self.password, charset=self.charset, cursorclass=pymysql.cursors.DictCursor)
                self.logger.info("...ok")
            except:
                rv = plugintypes.ConnectorErrors.FORBIDDEN
                self.logger.error("Error conencting to mysql server")
                self.logger.info("full traceback:\n%s"%traceback.format_exc())
        
        return rv


from openProduction.suite import TestSuite
from functools import wraps

def skipIf(_lambda):
    def wrapper(f):
        @wraps(f)
        def wrapped(self, *f_args, **f_kwargs):
            if callable(_lambda):
                if _lambda(self) == False:
                    return f(self, *f_args, **f_kwargs)
                else:
                    return TestSuite.ExecutionResult.SKIP
            else:
                if _lambda:
                    return TestSuite.ExecutionResult.SKIP
        return wrapped
    return wrapper


from openProduction.common import misc
from openProduction.connectors.BaseConnector import ConnectorErrors
import logging
from openProduction.connectors import ConnectionManager, BaseConnector
import traceback
import os
import json
import posixpath
from smb import SMBConnection

class ServerInterface:
    DB_REFIX = "db_op_"
    STATION_TABLE = "Stations"
    PRODUCT_TABLE = "Products"
    PRODUCT_REVISIONS_TABLE = "ProductRevisions"
    PRODUCT_STEPS_TABLE = "ProductSteps"
    PRODUCT_STATION_LINK_TABLE = "ProductStationLink"
    PRODUCT_RESULTS_TABLE = "ProductResults"
    GIT_CREDENTIALS_TABLE = "GitCredentials"
    SFTP_CONFIG_TABLE = "SFTPConfigTable"
    
    def __init__(self, workDir, connector, workspace):
        self.logger = logging.getLogger(misc.getAppName())
        self.workDir = workDir
        self.connector = connector
        self.workspace = workspace
        self.dbName = self._workspace2dbName(workspace)
        
    @staticmethod        
    def create(workDir):
        logger = logging.getLogger(misc.getAppName())
        cfgFile = os.path.abspath(os.path.join(workDir, "config.json"))
        
        logger.info("trying to open file %s"%os.path.abspath(cfgFile))
        try:
            f = open(cfgFile)
        except Exception as e:
            logger.error("opening config file %s failed with error message %s"%(cfgFile, str(e)))
            logger.info("full traceback:\n%s"%traceback.format_exc())
            return None
            
        logger.info("trying to parse content of config file to json")
        try:
            config = json.load(f)
        except Exception as e:
            logger.error("parsing config file %s failed with error message %s"%(cfgFile, str(e)))
            logger.info("full traceback:\n%s"%traceback.format_exc())
            return None
            
        connectorName = config["connectors"][0]["connectorName"]
        connectorCfg = config["connectors"][0]["connectorConfig"]        
        logger.info("trying to load connector '%s'"%connectorName)
        c = ConnectionManager.ConnectionManager()
        rv, connector = c.loadConnector(connectorName, connectorCfg, workDir)        
        if rv.value != BaseConnector.ConnectorErrors.NO_ERROR.value:
            logger.error("loading connector station failed with error %s"%str(rv))
            return None

        workspace = config["workspace"]        
        return ServerInterface(workDir, connector, workspace)
    
    def _workspace2dbName(self, workspace):
        return ServerInterface.DB_REFIX+workspace   
    
    ##########################################################################
    #                       git credentials manipulation                     #
    ##########################################################################
    def createGitCredentials(self, data):
        rv, _id = self.connector.addRow(self.dbName, self.GIT_CREDENTIALS_TABLE, data)
        if rv == ConnectorErrors.NO_ERROR:
            data["credential_id"] = _id
        return rv, data
            
    def deleteGitCredentials(self, credential_id):
        filt = "`credential_id` = %d"%credential_id
        return self.connector.deleteRows(self.dbName, self.GIT_CREDENTIALS_TABLE, filt)
    
    def listGitCredentials(self, filt=None):
        return self.connector.getData(self.dbName, self.GIT_CREDENTIALS_TABLE, ["*"], filt=filt)
    
    def updateGitCredentials(self, credential_id, data):
        filt = "`credential_id` = %d"%credential_id
        return self.connector.updateRow(self.dbName, self.GIT_CREDENTIALS_TABLE, data, filt)
    
    def getGitCredentials(self, credential_id):
        filt = "`credential_id` = %d"%credential_id
        rv, data = self.connector.getData(self.dbName, self.GIT_CREDENTIALS_TABLE, ["*"], filt=filt)
        if data == None:
            rv = ConnectorErrors.NOT_FOUND
        else:
            data = data[0]
        return rv, data    
    
    def _gitCredentialsTable(self):
        columns = [{
                "name": "git_credential_id",
                "type": "int",
                "primaryKey": True,
                "notNull": True,
                "autoIncrement": True,
            },
            {
                "name": "git_url",
                "type": "CHAR(255)",
                "unique": True,
                "notNull": True
            },
            {
                "name": "git_username",
                "type": "CHAR(255)",
                "notNull": True
            },
            {
                "name": "git_password",
                "type": "CHAR(255)",
                "notNull": True
            },
            {
                "name": "git_usermail",
                "type": "CHAR(255)",
                "notNull": True
            }
        ]
        table = {
            "name": self.GIT_CREDENTIALS_TABLE,
            "columns": columns
        }
        return table    
    
    
    ##########################################################################
    #                    product station link manipulation                   #
    ##########################################################################
    def createProductStationLink(self, data):
        rv, _id = self.connector.addRow(self.dbName, self.PRODUCT_STATION_LINK_TABLE, data)
        if rv == ConnectorErrors.NO_ERROR:
            data["link_id"] = _id
        return rv, data
            
    def deleteProductStationLink(self, link_id):
        filt = "`link_id` = %d"%link_id
        return self.connector.deleteRows(self.dbName, self.PRODUCT_STATION_LINK_TABLE, filt)
    
    def listProductStationLinks(self, filt=None, columns=["*"]):
        rv, data =  self.connector.getData(self.dbName, self.PRODUCT_STATION_LINK_TABLE, columns, filt=filt)
        if data == None:
            rv = ConnectorErrors.NOT_FOUND
        return rv, data
    
    def updateProductStationLink(self, link_id, data):
        filt = "`link_id` = %d"%link_id
        return self.connector.updateRow(self.dbName, self.PRODUCT_STATION_LINK_TABLE, data, filt)
    
    def getProductStationLink(self, link_id):
        filt = "`link_id` = %d"%link_id
        rv, data = self.connector.getData(self.dbName, self.PRODUCT_STATION_LINK_TABLE, ["*"], filt=filt)
        if data == None:
            rv = ConnectorErrors.NOT_FOUND
        else:
            data = data[0]
        return rv, data    
    
    def getPreviousStationID(self, station_id, product_id):
        filt = "`product_id`=%d and `station_id`=%d"%(product_id,station_id)
        prevStationID = None
        rv, data = self.listProductStationLinks(filt)
        if rv == ConnectorErrors.NO_ERROR:
            order = data[0]["order"]
            filt = "`product_id`=%d and `order`<%d"%(product_id,order)
            rv, data = self.listProductStationLinks(filt)            
            if rv == ConnectorErrors.NO_ERROR:
                maxOrder = -1
                for i in data:
                    if i["order"] > maxOrder:
                        maxOrder = i["order"]
                        prevStationID = i["station_id"]
        return rv, prevStationID
    
    def _productStationLinkTable(self):
        columns = [{
                "name": "link_id",
                "type": "int",
                "primaryKey": True,
                "notNull": True,
                "autoIncrement": True,
            },
            {
                "name": "product_id",
                "type": "int",
                "notNull": True,
                "foreign": [self.PRODUCT_TABLE, "product_id"]
            },
            {
                "name": "station_id",
                "type": "int",
                "notNull": True,
                "foreign": [self.STATION_TABLE, "station_id"]
            },
            {
                "name": "order",
                "type": "int",
                "notNull": True
            }
        ]
        table = {
            "name": self.PRODUCT_STATION_LINK_TABLE,
            "columns": columns,
            "uniqueKey": ["product_id", "order"]
        }
        return table    

    ##########################################################################
    #                      Product result manipulation                       #
    ##########################################################################
    def createProductResult(self, data):
        rv, _id = self.connector.addRow(self.dbName, self.PRODUCT_RESULTS_TABLE, data)
        if rv == ConnectorErrors.NO_ERROR:
            data["product_result_id"] = _id
        return rv, data
            
    def deleteProductResult(self, product_result_id):
        filt = "`product_result_id` = %d"%product_result_id
        return self.connector.deleteRows(self.dbName, self.PRODUCT_RESULTS_TABLE, filt)
    
    def listProductResults(self, filt=None, columns=["*"], joins=None):
        rv, data = self.connector.getData(self.dbName, self.PRODUCT_RESULTS_TABLE, columns, filt=filt, joins=joins)    
        if data == None:
            rv = ConnectorErrors.NOT_FOUND
        return rv, data   
    
    def getProductResult(self, product_result_id):
        filt = "`product_result_id` = %d"%product_result_id
        rv, data = self.connector.getData(self.dbName, self.PRODUCT_RESULTS_TABLE, ["*"], filt=filt)
        if data == None:
            rv = ConnectorErrors.NOT_FOUND
        else:
            data = data[0]
        return rv, data
    
    def _productResultsTable(self):
        columns = [{
                "name": "product_result_id",
                "type": "int",
                "primaryKey": True,
                "notNull": True,
                "autoIncrement": True,
            },
            {
                "name": "revision_id",
                "type": "int",
                "notNull": True,
                "foreign": [self.PRODUCT_REVISIONS_TABLE, "revision_id"]
            },
            {
                "name": "station_id",
                "type": "int",
                "notNull": True,
                "foreign": [self.STATION_TABLE, "station_id"]
            },
            {
                "name": "result",
                "type": "JSON",
                "notNull": True
            },
            {
                "name": "duration",
                "type": "FLOAT ",
                "notNull": True
            },
            {
                "name": "testCaseResults",
                "type": "JSON",
                "notNull": True
            },
                    {
                "name": "tempParams",
                "type": "JSON",
                "notNull": True
            },
            {
                "name": "openProductionVersion",
                "type": "CHAR(11)",
                "notNull": True
            },
            {
                "name": "success",
                "type": "bool",
                "notNull": True
            },
            {
                "name": "date",
                "type": "DATETIME",
                "constraint": "DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP"
            },
            {
                "name": "deviceID",
                "type": "CHAR(40)",
                "notNull": True
            },
        ]
        table = {
            "name": self.PRODUCT_RESULTS_TABLE,
            "columns": columns
        }
        return table
    
    ##########################################################################
    #                       Product steps manipulation                       #
    ##########################################################################
    def createProductStep(self, data, revisionData):
        stepData = None
        rv = self.connector.startTransaction(self.dbName)
        if rv == ConnectorErrors.NO_ERROR:
            # add to product table
            rv, _id = self.connector.addRow(self.dbName, self.PRODUCT_STEPS_TABLE, data)
            
            if rv == ConnectorErrors.NO_ERROR:
                stepData = data
                stepData["product_step_id"] = _id
                revisionData["product_step_id"] = _id
                rv, _id = self.createProductRevision(revisionData)
                if rv == ConnectorErrors.NO_ERROR:
                    revisionData["revision_id"] = _id
                            
            if rv != ConnectorErrors.NO_ERROR:
                self.connector.rollbackTransaction()
                stepData = None
                revisionData = None
            else:
                rv = self.connector.commitTransaction()
        else:
            revisionData = None
        
        return rv, stepData, revisionData
        
        
        rv, _id = self.connector.addRow(self.dbName, self.PRODUCT_STEPS_TABLE, data)        
        if rv == ConnectorErrors.NO_ERROR:
            data["product_step_id"] = _id
        return rv, data
            
    def deleteProductStep(self, product_step_id):
        filt = "`product_step_id` = %d"%product_step_id
        return self.connector.deleteRows(self.dbName, self.PRODUCT_STEPS_TABLE, filt)
    
    def listProductSteps(self, filt=None):
        return self.connector.getData(self.dbName, self.PRODUCT_STEPS_TABLE, ["*"], filt=filt)
    
    def updateProductStep(self,  product_step_id, data):
        filt = "`product_step_id` = %d"%product_step_id
        return self.connector.updateRow(self.dbName, self.PRODUCT_STEPS_TABLE, data, filt)
    
    def getProductStep(self, product_step_id):
        filt = "`product_step_id` = %d"%product_step_id
        rv, data = self.connector.getData(self.dbName, self.PRODUCT_STEPS_TABLE, ["*"], filt=filt)
        if data == None:
            rv = ConnectorErrors.NOT_FOUND
        else:
            data = data[0]
        return rv, data

    def getProductStepByName(self, stationName, productName, versionName):
        filt = "prod.`name` like '%s' and station.`name` like '%s' and version like '%s'"%(productName, stationName, versionName)
        
        joins = ["`%s`.`Products` prod ON prod.product_id = ProductSteps.product_id"%self.dbName,
                 "`%s`.`Stations` station ON station.station_id = ProductSteps.station_id"%self.dbName,
                 "`%s`.`GitCredentials` cred ON cred.git_credential_id = ProductSteps.git_credential_id"%self.dbName
                 ]
        
        rv, data = self.connector.getData(self.dbName, self.PRODUCT_STEPS_TABLE, ["*"], filt=filt, joins=joins)
        if data == None:
            rv = ConnectorErrors.NOT_FOUND
        else:
            data = data[0]
        return rv, data
    
    def _productStepsTable(self):
        columns = [{
                "name": "product_step_id",
                "type": "int",
                "primaryKey": True,
                "notNull": True,
                "autoIncrement": True,
            },
            {
                "name": "product_id",
                "type": "int",
                "notNull": True,
                "foreign": [self.PRODUCT_TABLE, "product_id"]
            },
            {
                "name": "version",
                "type": "CHAR(255)",
                "notNull": True
            },
            {
                "name": "git_credential_id",
                "type": "int",
                "notNull": True,
                "foreign": [self.GIT_CREDENTIALS_TABLE, "git_credential_id"]
            },
            {
                "name": "git_branch",
                "type": "CHAR(255)",
                "notNull": True
            },
            {
                "name": "description",
                "type": "CHAR(255)",
                "notNull": True
            },
            {
                "name": "image",
                "type": "MEDIUMBLOB",
                "notNull": True
            },
            {
                "name": "state",
                "type": "int",
                "default": 0,
                "notNull": True
            },
            {
                "name": "station_id",
                "type": "int",
                "notNull": True,
                "foreign": [self.STATION_TABLE, "station_id"]
            }
        ]
        table = {
            "name": self.PRODUCT_STEPS_TABLE,
            "columns": columns,
            "uniqueKey": ["product_id", "version", "station_id"]
        }
        return table
    
    ##########################################################################
    #                      product revision manipulation                     #
    ##########################################################################
    def createProductRevision(self, data):
        if "product_step_id" not in data:
            return ConnectorErrors.INVALID_PARAM, None
        else:
            rv, rvData = self.getLatestProductRevision(data["product_step_id"])
            if rv != ConnectorErrors.NOT_FOUND:
                return ConnectorErrors.EXISTS, None
        
        rv, _id = self.connector.addRow(self.dbName, self.PRODUCT_REVISIONS_TABLE, data)
        if rv == ConnectorErrors.NO_ERROR:
            data["revision_id"] = _id
        return rv, data
            
    def deleteProductRevision(self, revision_id):
        filt = "`revision_id` = %d"%revision_id
        return self.connector.deleteRows(self.dbName, self.PRODUCT_REVISIONS_TABLE, filt)
    
    def listProductRevisions(self, filt=None, joins=None):
        return self.connector.getData(self.dbName, self.PRODUCT_REVISIONS_TABLE, ["*"], filt=filt, joins=joins, orderBy="revision_id DESC")
    
    def getLatestProductRevisionByName(self, stationName, productName, versionName):
        rv, rvData = self.getProductStepByName(stationName, productName, versionName)
        if rv == ConnectorErrors.NO_ERROR:
            stepId = rvData["product_step_id"]
            rv, rvData = self.getLatestProductRevision(stepId)
        else:
            rvData = None
        return rv, rvData    
    
    def getLatestProductRevision(self, product_step_id):
        filt = "`product_step_id` = %d and revision_id=(select max(revision_id) from `%s`.`%s` WHERE `product_step_id` = %d)"%(product_step_id, self.dbName, self.PRODUCT_REVISIONS_TABLE, product_step_id)
        rv, rvData = self.connector.getData(self.dbName, self.PRODUCT_REVISIONS_TABLE, ["*"], filt=filt)
        if rvData == None:
            rv = ConnectorErrors.NOT_FOUND
        else:
            rvData = rvData[0]
            rvData["params"] = json.loads(rvData["params"])
        return rv, rvData
    
    def updateProductRevision(self, product_step_id, data):
        entry = None
        if "product_step_id" in data:
            data.pop("product_step_id", None)    
        
        rv, rvData = self.getLatestProductRevision(product_step_id)
        if rv == ConnectorErrors.NO_ERROR:
            entry = rvData
            entry.pop("revision_id", None)
            entry.pop("date", None)
            for key, val in data.items():
                entry[key] = val
            rv, _id = self.connector.addRow(self.dbName, self.PRODUCT_REVISIONS_TABLE, entry)
            entry["revision_id"] = _id
        
        return rv, entry
    
    def updateProductRevisionByName(self, stationName, productName, versionName, data):
        rv, rvData = self.getProductStepByName(stationName, productName, versionName)
        if rv == ConnectorErrors.NO_ERROR:
            stepId = rvData["product_step_id"]
            rv, rvData = self.updateProductRevision(stepId, data)
        else:
            rvData = None
        return rv, rvData
        
    def getProductRevision(self, revision_id):
        filt = "`revision_id` = %d"%revision_id
        rv, data = self.connector.getData(self.dbName, self.PRODUCT_REVISIONS_TABLE, ["*"], filt=filt)
        if data == None:
            rv = ConnectorErrors.NOT_FOUND
        else:
            data = data[0]
            data["params"] = json.loads(data["params"])
        return rv, data
    
    def _productRevisionTable(self):
        columns = [{
                "name": "revision_id",
                "type": "int",
                "primaryKey": True,
                "notNull": True,
                "autoIncrement": True,
            },
            {
                "name": "product_step_id",
                "type": "int",
                "notNull": True,
                "foreign": [self.PRODUCT_STEPS_TABLE, "product_step_id"]
                
            },
            {
                "name": "params",
                "type": "JSON",
                "notNull": True
            },            
            {
                "name": "commit_id",
                "type": "CHAR(40)",
                "notNull": True
            },
            {
                "name": "date",
                "type": "DATETIME",
                "constraint": "DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP"
            }
        ]
        table = {
            "name": self.PRODUCT_REVISIONS_TABLE,
            "columns": columns
        }
        return table    

    ##########################################################################
    #                             product manipulation                       #
    ##########################################################################
    def createProduct(self, data, stationLinks):
        """stationLinks: list of dicts
        """
        productData = None
        linkData = None
        rv = self.connector.startTransaction(self.dbName)
        if rv == ConnectorErrors.NO_ERROR:
            # add to product table
            rv, _id = self.connector.addRow(self.dbName, self.PRODUCT_TABLE, data)
            
            if rv == ConnectorErrors.NO_ERROR:
                productData = data
                productData["product_id"] = _id
                linkData = []
                
                # add to station link table
                if type(stationLinks) == type({}):
                    stationLinks = [stationLinks]
                
                for link in stationLinks:
                    link["product_id"] = _id
                    rv, stationLink = self.createProductStationLink(link)
                    if rv == ConnectorErrors.NO_ERROR:
                        linkData.append(stationLink)
                    else:
                        break
            
            if rv != ConnectorErrors.NO_ERROR:
                self.connector.rollbackTransaction()
                productData = None
                linkData = None
            else:
                rv = self.connector.commitTransaction()
        
        return rv, productData, linkData
            
    def deleteProduct(self, product_id):
        filt = "`product_id` = %d"%product_id
        return self.connector.deleteRows(self.dbName, self.PRODUCT_TABLE, filt)
    
    def listProducts(self, filt=None):
        return self.connector.getData(self.dbName, self.PRODUCT_TABLE, ["*"], filt=filt)
    
    def updateProduct(self, product_id, data):
        filt = "`product_id` = %d"%product_id
        return self.connector.updateRow(self.dbName, self.PRODUCT_TABLE, data, filt)
    
    def getProduct(self, product_id):
        filt = "`product_id` = %d"%product_id
        rv, data = self.connector.getData(self.dbName, self.PRODUCT_TABLE, ["*"], filt=filt)
        if data == None:
            rv = ConnectorErrors.NOT_FOUND
        else:
            data = data[0]
        return rv, data    
    
    def _productTable(self):
        columns = [{
                "name": "product_id",
                "type": "int",
                "primaryKey": True,
                "notNull": True,
                "autoIncrement": True,
            },
            {
                "name": "name",
                "type": "CHAR(255)",
                "notNull": True,
                "unique": True,
            },
            {
                "name": "description",
                "type": "CHAR(255)",
                "notNull": True
            },
            {
                "name": "state",
                "type": "int",
                "default": 0,
                "notNull": True
            }
        ]
        table = {
            "name": self.PRODUCT_TABLE,
            "columns": columns
        }
        return table

    ##########################################################################
    #                             station manipulation                       #
    ##########################################################################
    def createStation(self, data):
        rv, station_id = self.connector.addRow(self.dbName, self.STATION_TABLE, data)
        if rv == ConnectorErrors.NO_ERROR:
            data["station_id"] = station_id
        return rv, data
            
    def deleteStation(self, station_id):
        filt = "`station_id` = %d"%station_id
        return self.connector.deleteRows(self.dbName, self.STATION_TABLE, filt)
    
    def listStations(self, filt=None):
        return self.connector.getData(self.dbName, self.STATION_TABLE, ["*"], filt=filt)
    
    def updateStation(self, station_id, data):
        filt = "`station_id` = %d"%station_id
        return self.connector.updateRow(self.dbName, self.STATION_TABLE, data, filt)
    
    def getStation(self, station_id):
        filt = "`station_id` = %d"%station_id
        rv, data = self.connector.getData(self.dbName, self.STATION_TABLE, ["*"], filt=filt)
        if data == None:
            rv = ConnectorErrors.NOT_FOUND
        else:
            data = data[0]
        return rv, data
    
    def listStationProducts(self, station_id):
        filt = "`station_id` = %d"%station_id
        joins = ["`%s`.`Products` prod ON prod.product_id = ProductStationLink.product_id"%self.dbName]
        return self.connector.getData(self.dbName, self.PRODUCT_STATION_LINK_TABLE, ["prod.*"], filt=filt, joins=joins)

    def listStationProductSteps(self, station_id, product_id):
        filt = "steps.`station_id` = %d and steps.`product_id` = %d"%(station_id, product_id)
        joins = ["`%s`.`ProductSteps` steps ON (steps.`product_id` = ProductStationLink.`product_id` and \
steps.`station_id` = ProductStationLink.`station_id`)"%self.dbName]
        return self.connector.getData(self.dbName, self.PRODUCT_STATION_LINK_TABLE, ["steps.*"], filt=filt, joins=joins)
    
    def _stationTable(self):
        columns = [{
                "name": "station_id",
                "type": "int",
                "primaryKey": True,
                "notNull": True,
                "autoIncrement": True,
            },
            {
                "name": "name",
                "type": "CHAR(255)",
                "notNull": True,
                "unique": True,
            },
            {
                "name": "description",
                "type": "CHAR(255)",
                "notNull": True
            }
        ]
        table = {
            "name": self.STATION_TABLE,
            "columns": columns
        }
        return table
    
    ##########################################################################
    #                         SFTP config manipulation                       #
    ##########################################################################
    def createSFTPConfig(self, data):
        rv, sftp_config_id = self.connector.addRow(self.dbName, self.SFTP_CONFIG_TABLE, data)
        if rv == ConnectorErrors.NO_ERROR:
            data["sftp_config_id"] = sftp_config_id
        return rv, data
            
    def deleteSFTPConfig(self, sftp_config_id):
        filt = "`sftp_config_id` = %d"%sftp_config_id
        return self.connector.deleteRows(self.dbName, self.SFTP_CONFIG_TABLE, filt)
    
    def listSFTPConfig(self, filt=None, columns=["*"], joins=None):
        rv, data = self.connector.getData(self.dbName, self.SFTP_CONFIG_TABLE, columns, filt=filt, joins=joins)    
        if data == None:
            rv = ConnectorErrors.NOT_FOUND
        return rv, data   
    
    def updateSFTPConfig(self, sftp_config_id, data):
        filt = "`sftp_config_id` = %d"%sftp_config_id
        return self.connector.updateRow(self.dbName, self.SFTP_CONFIG_TABLE, data, filt)
    
    def getSFTPConfig(self, sftp_config_id):
        filt = "`sftp_config_id` = %d"%sftp_config_id
        rv, data = self.connector.getData(self.dbName, self.SFTP_CONFIG_TABLE, ["*"], filt=filt)
        if data == None:
            rv = ConnectorErrors.NOT_FOUND
        else:
            data = data[0]
        return rv, data
    
    def _sftpConfigTable(self):
        columns = [{
                "name": "sftp_config_id",
                "type": "int",
                "primaryKey": True,
                "notNull": True,
                "autoIncrement": True,
            },
            {
                "name": "hostname",
                "type": "CHAR(255)",
                "notNull": True
            },
            {
                "name": "user",
                "type": "CHAR(255)",
                "notNull": True
            },
            {
                "name": "password",
                "type": "CHAR(255)",
                "notNull": True
            },
                    {
                "name": "workdir",
                "type": "CHAR(255)",
                "notNull": True
            }
        ]
        table = {
            "name": self.SFTP_CONFIG_TABLE,
            "columns": columns
        }
        return table
    
    ##########################################################################
    #                     Product result file manipulation                   #
    ##########################################################################
    def _getSMBObj(self, data):
        if hasattr(self, "smb") == False:
            user = data["user"]
            pw = data["password"]
            domain = ""
            system_name = data["hostname"]
            self.smb = SMBConnection.SMBConnection(user,pw,'name',system_name,domain,use_ntlm_v2=True,
                 is_direct_tcp=True) 
            connected = self.smb.connect(system_name,445)
            if connected == False:
                self.logger.error("Not connected over SMB")
                return None
        
        return self.smb
    
    def createProductResultFile(self, data):
        result_id = data["result_id"]
        localFile = data["filename"]
        sftp_config_id = data["sftp_config_id"]
        rv, data = self.getSFTPConfig(sftp_config_id)
        if rv == ConnectorErrors.NO_ERROR:
            smb = self._getSMBObj(data)
            if smb == None:
                return ConnectorErrors.UNSPECIFIC
            
            try:
                resultFolder = "result_id_%d"%result_id
                smb.createDirectory(data["workdir"], resultFolder)
            except:
                pass
            
            fName = os.path.split(localFile)[-1]
            
            f = open(localFile, "rb")
            resultFile = posixpath.join(resultFolder, fName)
            smb.storeFile(data["workdir"], resultFile, f)
            f.close()

        return rv
    
    def listProductResultFiles(self, result_id, sftp_config_id):
        files = None
        rv, data = self.getSFTPConfig(sftp_config_id)
        if rv == ConnectorErrors.NO_ERROR:
            smb = self._getSMBObj(data)
            if smb == None:
                return ConnectorErrors.UNSPECIFIC, []
            
            resultFolder = "result_id_%d"%result_id
            files = smb.listPath(data["workdir"], resultFolder)
            files = [x.filename for x in files]
            files.remove(".")
            files.remove("..")
                    
        return rv, files
        
    def downloadProductResultFile(self, result_id, sftp_config_id, remoteFileName, localFileName):
        rv, data = self.getSFTPConfig(sftp_config_id)
        if rv == ConnectorErrors.NO_ERROR:
            smb = self._getSMBObj(data)
            if smb == None:
                return ConnectorErrors.UNSPECIFIC
            
            remoteFolder = "result_id_%d"%result_id
            remoteFile = posixpath.join(remoteFolder, remoteFileName)
            localFP = open(localFileName, "wb")
            smb.retrieveFile(data["workdir"], remoteFile, localFP)
            localFP.close()            
        return rv
        
    ##########################################################################
    #                            workspace manipulation                      #
    ##########################################################################
    def createWorkspace(self, workspace):
        if workspace in self.listWorkspaces():
            return ConnectorErrors.EXISTS
        
        dbName = self._workspace2dbName(workspace)
        
        rv = self.connector.createDatabase(dbName)
        if rv == ConnectorErrors.NO_ERROR:
            tables = []
            tables.append(self._stationTable())
            tables.append(self._gitCredentialsTable())
            tables.append(self._productTable())
            tables.append(self._productStepsTable())
            tables.append(self._productRevisionTable())
            tables.append(self._productResultsTable())
            tables.append(self._productStationLinkTable())
            rv = self.connector.createTables(self.dbName, tables)
        else:
            return rv
        
        return rv
    
    def deleteWorkspace(self, workspace):
        dbName = self._workspace2dbName(workspace)
        return self.connector.deleteDatabase(dbName)
    
    def listWorkspaces(self):
        rv = []
        dbs = self.connector.listDatabases()
        for db in dbs:
            if db["Database"].find(ServerInterface.DB_REFIX) == 0:
                rv.append(db["Database"][len(ServerInterface.DB_REFIX):])
        return rv



from yapsy.PluginManager import PluginManager
import os
from openProduction.common import misc
import logging
from openProduction.connectors import BaseConnector

class ConnectionManager:
    def __init__(self):
        self.logger = logging.getLogger(misc.getAppName())
        self._manager = PluginManager()
        self._manager.setPluginPlaces([os.path.abspath(os.path.dirname(__file__))])
        self._manager.setCategoriesFilter({
                "Connectors" : BaseConnector.BaseConnector})
        
        self._manager.collectPlugins()
    
        # Loop round the plugins and print their names.
        self._connectors = []
        
        self.logger.info("loading connector plugins now")
#        for plugin in self._manager.getAllPlugins():
        for plugin in self._manager.getPluginsOfCategory("Connectors"):
            self.logger.info("found connector '%s'"%plugin.name)
            self._connectors.append(plugin.name)
            
        self._connectors = sorted(self._connectors)
        self.logger.info("... found %d connector(s)"%len(self._connectors))
    
    def getConnectors(self):
        return self._connectors
    
    def getConnectorConfigDesc(self, name):
        cfg = []
        connector = self._manager.activatePluginByName(name, category="Connectors")
        if connector != None:
            cfg = connector.getConfig()
        return cfg
    
    def loadConnector(self, name, config, workdir):
        rv = BaseConnector.ConnectorErrors.NO_ERROR
        connector = self._manager.activatePluginByName(name, category="Connectors")
        if connector != None:
            rv = connector.applyConfig(config, workdir)
        return rv, connector


# -*- coding: utf-8 -*-
"""
Created on Thu Apr 25 15:41:59 2019

@author: Markus
"""
from openProduction.common.Types import String, Float, Bool, BaseType, Date

class TestSuiteValues:
    
    def __init__(self):
        self._entries = []
        self.addValue(String("OPEN_PRODUCTION_VERSION", "Version string of open prodcution"))
        self.addValue(String("DEVICE_UUID", "Global Unique ID of DUT"))
        self.addValue(Date("DATE", "Run date of device"))
        self.addValue(String("VERSION", "Version string of testsuite"))
        self.addValue(Float("DURATION", "Duration of test run in seconds"))
        self.addValue(Bool("SUCCESS", "True if product run was successful in this station"))
    
    def addValue(self, value, errorOnExist=True):
        if isinstance(value, BaseType) == False:
            raise RuntimeError("value has to be a subclass of openProduction.common.Types.BaseType")
        
        exists = False
        
        for e in self._entries:
            if e.name == value.name:
                exists = True
            
        if exists and errorOnExist:
            raise RuntimeError("value %s already exists"%value.name)
            
        if exists == False:
            setattr(self, value.name, value)
            self._entries.append(value)
        
    def addValueContainer(self, container):
        if isinstance(container, TestSuiteValues) == False:
            raise RuntimeError("container has to be a subclass of openProduction.suite.TestSuiteValueContainer.TestSuiteValues")
            
        for val in container:
            self.addValue(val, errorOnExist=False)
    
    def getNames(self):
        names = []
        for e in self._entries:
            names.append(e.name)
        return names
    
    def getValue(self, name):
        for e in self._entries:
            if e.name == name:
                return e.getValue()
            
        raise RuntimeError("value %s not found"%name)
    
    def setValue(self, name, value):
        for e in self._entries:
            if e.name == name:
                e.setValue(value)
                return
            
        raise RuntimeError("value %s not found"%name)
    
    def clear(self):
        for val in self._entries:
            val.clear()
    
    def __setattr__(self, name, value):
        try:
            for e in self._entries:
                if e.name == name:
                    e.setValue(value)
                    return
        except:
            pass
        
        self.__dict__[name] = value
    
    def __len__(self):
        return len(self._entries)
        
    def __iter__(self):
        self.idx = 0
        return self
        
    def __next__(self):
        if self.idx >= len(self._entries):
            raise StopIteration
        else:
            self.idx += 1
            return self._entries[self.idx-1]
        
    def __str__(self):
        maxName = len("Name")
        maxValue = len("Value")
        
        for e in self._entries:
            maxName = max(maxName, len(e.name))
            maxValue = max(maxValue, len(str(e.value)))
            
        colName = maxName+2
        colVal = maxValue+2
        
        myStr = ""
        myStr += ("+"+"-"*colName + "+" + "-"*colVal + "+") + "\n"
        myStr += ("| Name " + " "*(colName-6) + "| Value " + " "*(colVal-7) + "|") + "\n"
        myStr += ("+"+"="*colName + "+" + "="*colVal + "+") + "\n"
        
        for e in self._entries:
            myStr += ("| " + e.name + " "*(colName-len(e.name)-1) + "| " +
                      str(e.value) +" "*(colVal-len(str(e.value))-1) + "|" ) + "\n"
            myStr += ("+"+"-"*colName + "+" + "-"*colVal + "+") + "\n"
        return myStr


# -*- coding: utf-8 -*-
"""
Created on Sun Jun 23 16:33:09 2019

@author: Markus
"""
import pandas as pd
from openProduction.connectors.BaseConnector import ConnectorErrors
from openProduction.server import ServerInterface

class Report:
    def __init__(self, serverIF):
        self.serverIF = serverIF
        
    @staticmethod
    def create(workDir):
        return Report(ServerInterface.ServerInterface.create(workDir))
    
    def getNumDevices(self, productName, stationName):
        pass
    
    def getResultKeys(self, productName, productVersion, stationName, onlyLatestRev=False, specificRev=None):
        rv = ConnectorErrors.NO_ERROR
        columns = ["JSON_KEYS(`result`) AS 'keys'"]
        data = None
        
        if specificRev != None:
            filt = ["ProductResults.`revision_id`=%d"%specificRev]
        elif onlyLatestRev:
            rv, rev = self.serverIF.getLatestProductRevisionByName(stationName, productName, productVersion)
            if rv == ConnectorErrors.NO_ERROR:
                revId= rev["revision_id"]
                filt = ["ProductResults.`revision_id`=%d"%revId]
        else:
            filt = []
            if stationName != None:
                filt += ["station.`name` like '%s'"%stationName]
            if productName != None:
                filt += ["prod.`name` like '%s'"%productName]
            if productVersion != None:
                filt += ["steps.`version` like '%s'"%productVersion]
            
        if rv == ConnectorErrors.NO_ERROR:
            filt += ["`success`=1"]
            filt = " and ".join(filt)
            
            joins = [
                     "`%s`.`Stations` station ON station.station_id = ProductResults.station_id"%self.serverIF.dbName,
                     "`%s`.`ProductRevisions` rev ON rev.revision_id = ProductResults.revision_id"%self.serverIF.dbName,
                     "`%s`.`ProductSteps` steps ON steps.product_step_id = rev.product_step_id"%self.serverIF.dbName,
                     "`%s`.`Products` prod ON steps.product_id = prod.product_id"%self.serverIF.dbName,
                     ]            
            rv, data = self.serverIF.listProductResults(filt=filt, columns=columns, joins=joins)
            
            if rv == ConnectorErrors.NO_ERROR:
                if onlyLatestRev or specificRev != None:
                    data = data[0]
            
        return rv, data
    
    def getDeviceResults(self, productName, productVersion, stationName, deviceID=None, resultKeys=None, onlyGood=True):
        columns = ["`deviceID`", "`duration`", "ProductResults.`date` as date", "ProductResults.`revision_id` as revision_id"]
        
        filt = []
        if stationName != None:
            filt += ["station.`name` like '%s'"%stationName]
        if productName != None:
            filt += ["prod.`name` like '%s'"%productName]
        if productVersion != None:
            filt += ["steps.`version` like '%s'"%productVersion]
        if deviceID != None:
            filt += ["`deviceID` like '%s'"%deviceID]
        
        if onlyGood:
            filt += ["`success`=1"]
        
        columns += ["success"]
            
        filt = " and ".join(filt)
        
        joins = [
                 "`%s`.`Stations` station ON station.station_id = ProductResults.station_id"%self.serverIF.dbName,
                 "`%s`.`ProductRevisions` rev ON rev.revision_id = ProductResults.revision_id"%self.serverIF.dbName,
                 "`%s`.`ProductSteps` steps ON steps.product_step_id = rev.product_step_id"%self.serverIF.dbName,
                 "`%s`.`Products` prod ON steps.product_id = prod.product_id"%self.serverIF.dbName,
                 ]
        
        
        if resultKeys != None:
            for k in resultKeys:
                if k.find("UNQUOTE_") == 0:
                    k = k[len("UNQUOTE_"):]
                    columns.append("JSON_UNQUOTE(JSON_EXTRACT(`result`, '$.%s')) as %s"%(k,k))
                else:
                    columns.append("JSON_EXTRACT(`result`, '$.%s') as %s"%(k,k))
        
        rv, data = self.serverIF.listProductResults(filt=filt, columns=columns, joins=joins)
        if rv == ConnectorErrors.NO_ERROR:
            p = pd.DataFrame(data)
            if resultKeys != None:
                for k in resultKeys:
                    if k.find("UNQUOTE_") == 0:
                        continue
                    try:
                        p[k] = pd.to_numeric(p[k])
                    except:
                        pass
        else:
            p = None
        return p
    
if __name__ == "__main__":
    from openProduction.common import misc
    from pylab import *
    
    rep = Report.create(misc.getDefaultAppFolder())
    rv = rep.getResultKeys("Bauernstolz", "AA", "Klebestation", onlyLatestRev=True)


from openProduction.common import misc
from openProduction.product import Product
import logging
import traceback

class ProductInstaller:
    
    def __init__(self):
        self.logger = logging.getLogger(misc.getAppName())
    
    def run(self, station, productName=None, productVersion=None):
        print ("Welcome to the product CLI installer")
        if productName == None:
            productName = misc.queryString("Please give a productName: ")
            productVersion = misc.queryString("Please give a productVersion: ")
        elif productVersion == None:
            productVersion = misc.queryString("Please give a productVersion: ")
            
        executor = misc.queryFile("Please specify the location of the station executor script: ")
        params = misc.queryFile("Please specify the location of the parameter file: ")
        
        revision = station.scm.getNextRevision(productName, productVersion)
        
        self.logger.info("Installing product %s with version %s. \
Revision for this product is %d"%(productName, productVersion, revision))
        
        product = Product(productName, productVersion, revision, executor, params)
        
        try:
            product.save(station)
            self.logger.info("...ok")
            ok = True
        except:
            self.logger.error("error adding product %s.%s.%d"%(productName,
                                                          productVersion,
                                                          revision))
            self.logger.info("full traceback:\n%s"%traceback.format_exc())
            ok = False
            
        return ok


import unittest
from openProduction.suite import TestClass, TestLoader

class TCCorrect(TestClass.TestClass):
    """doc class"""

    def __init__(self):
        pass

    def step__00100(self):
        """Step 1 example"""
        pass

    def step__00200(self):
        """Step 2 example"""
        pass

class TCNoBase:
    """doc class"""

    def __init__(self):
        pass

    def step__00100(self):
        """Step 1 example"""
        pass



class TestSuiteTests(unittest.TestCase):

    def test_correct_class(self):
        """Test if loading a correct inherited class works"""
        tc = TCCorrect()
        a = TestLoader.TestLoaderClass(tc)
        ts = a.discover()
        self.assertEqual(len(ts.getTestCases()), 2)
        
    def test_no_base_class(self):
        """Test if loading a test class without TestClass as base fails"""
        tc = TCNoBase()
        self.assertRaises(RuntimeError, TestLoader.TestLoaderClass, tc)
        
if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
"""
Created on Fri May 31 11:29:17 2019

@author: Markus
"""
from openProduction.suite import TestSuiteIO, TestSuite, TestRunnerCli
from concurrent.futures import ThreadPoolExecutor
from PyQt5 import QtCore, QtQml, QtMultimedia, QtGui, QtQuick
from openProduction.station import Station
from openProduction.common import misc
from openProduction import log
import os
import logging
import subprocess
import base64
import json
import threading
import functools
import queue

class IOHandlerQML(QtCore.QObject, TestSuiteIO.BaseIOHandler):
    
    queryYesNoSignal = QtCore.pyqtSignal(str, str, name='queryYesNoSignal')
    newImageDataGUI = QtCore.pyqtSignal(name='newImageDataGUI')
    
    def __init__(self):
        QtCore.QObject.__init__(self)
        TestSuiteIO.BaseIOHandler.__init__(self)
        self._syncEvt = threading.Event()
        self._imgs = []
        
    def newImageData(self, image, imageFormat):
        prov = MyImageProvider.getInstance()
        if imageFormat == TestSuiteIO.ImageFormat.GRAYSCALE_8:
            dtype = QtGui.QImage.Format_Grayscale8
        if imageFormat == TestSuiteIO.ImageFormat.RGB_888:
            dtype = QtGui.QImage.Format_RGB888

        qimage = QtGui.QImage(image, image.shape[1],image.shape[0], dtype)
        prov.pushImage(qimage)
        self.newImageDataGUI.emit()
        
    def _queryYesNo(self, title, msg, timeout=None):
        self._syncEvt.clear()
        self.queryYesNoSignal.emit(title, msg)
        rv = self._syncEvt.wait(timeout)
        if not rv:
            rv = TestSuiteIO.IOAnswer.TIMEOUT
        else:
            rv = self._yesNoAnswer
        return rv

    @QtCore.pyqtSlot(TestSuiteIO.IOAnswer)
    def answerYesNo(self, answer):
        self._yesNoAnswer = answer
        self._syncEvt.set()
    

class RevisionModel(QtCore.QAbstractListModel):

    IDX_NAME = QtCore.Qt.UserRole
    IDX_DATE = QtCore.Qt.UserRole+1

    def __init__(self, parent=None):
        super(RevisionModel, self).__init__(parent)
        self.items = []

    def add_item(self, item):
        self.items.append(item)

    def clearModel(self):
        #call self.endResetModel() after finishing with model
        self.beginResetModel()
        self.items = []

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self.items)

    @QtCore.pyqtSlot(int, result=str)
    def getName(self, index):
        try:
            return str(self.items[index][self.IDX_NAME])
        except IndexError:
            return ""

    @QtCore.pyqtSlot(int, result=str)
    def getDate(self, index):
        try:
            return self.items[index][self.IDX_DATE]
        except IndexError:
            return ""

    def data(self, index, role=QtCore.Qt.DisplayRole):
        try:
            rv =  self.items[index.row()][role]
            return rv
        except IndexError:
            return QtCore.QVariant()

    def roleNames(self):
        return {
            self.IDX_NAME: QtCore.QByteArray(b"name"),
            self.IDX_DATE: QtCore.QByteArray(b"date")
        }

class VersionModel(QtCore.QAbstractListModel):

    IDX_NAME = QtCore.Qt.UserRole
    IDX_COMMENT = QtCore.Qt.UserRole+1
    IDX_IMAGE = QtCore.Qt.UserRole+2

    def __init__(self, parent=None):
        super(VersionModel, self).__init__(parent)
        self.items = []

    def add_item(self, item):
        self.beginInsertRows(QtCore.QModelIndex(),
                             self.rowCount(),
                             self.rowCount())

        self.items.append(item)
        self.endInsertRows()

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self.items)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        try:
            return self.items[index.row()][role]
        except IndexError:
            return QtCore.QVariant()

    def roleNames(self):
        return {
            self.IDX_NAME: QtCore.QByteArray(b"name"),
            self.IDX_COMMENT: QtCore.QByteArray(b"comment"),
            self.IDX_IMAGE: QtCore.QByteArray(b"imageSource")
        }

class ProductModel(QtCore.QAbstractListModel):

    IDX_NAME = QtCore.Qt.UserRole
    IDX_DETAILS = QtCore.Qt.UserRole+1

    def __init__(self, parent=None):
        super(ProductModel, self).__init__(parent)
        self.items = []

    def add_item(self, item):
        #self.beginInsertRows(QtCore.QModelIndex(),
        #                     self.rowCount(),
        #                     self.rowCount())

        self.items.append(item)
        #self.endInsertRows()

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self.items)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        try:
            return self.items[index.row()][role]
        except IndexError:
            return QtCore.QVariant()

    def roleNames(self):
        return {
            self.IDX_NAME: QtCore.QByteArray(b"name"),
            self.IDX_DETAILS: QtCore.QByteArray(b"details")
        }

class Helper(QtCore.QObject):
    def __init__(self):
        super(Helper,self).__init__()

    @QtCore.pyqtSlot(str)
    def openExplorer(self, folderName):
        subprocess.Popen(r'explorer /select,"%s"'%folderName)

class MyLogger(logging.Handler, QtCore.QObject):

    newLogMessage = QtCore.pyqtSignal(str, name='newLogMessage')

    def __init__(self):
        logging.Handler.__init__(self)
        QtCore.QObject.__init__(self)

        self.txt = ""

    @QtCore.pyqtSlot()
    def resetLog(self):
        self.txt = ""

    @QtCore.pyqtSlot(result=str)
    def getErrorLog(self):
        return self.txt

    def emit(self, record):
        if record.levelno < logging.WARNING:
            self.newLogMessage.emit(record.msg)
        self.txt += self.format(record) + "\n"

class ParamsQML(QtCore.QAbstractListModel):
    IDX_KEY = QtCore.Qt.UserRole
    IDX_VALUE = QtCore.Qt.UserRole+1
    IDX_TYPE = QtCore.Qt.UserRole+2

    keyChanged = QtCore.pyqtSignal(str, object, name='keyChanged')
    
    def __init__(self, params):
        super(ParamsQML,self).__init__()
        self.params = params
        self.params.keyChanged.connect(self._onKeyChangedAsync)
        self.keyChanged.connect(self._onKeyChanged)
        self.syncModel()
        
    def _onKeyChangedAsync(self, key, value):
        self.keyChanged.emit(key, value)
        
    @QtCore.pyqtSlot(str, object)
    def _onKeyChanged(self, key, value):
        self.syncModel()
    
    @QtCore.pyqtSlot(str, QtCore.QVariant, str)
    def keyChangedFromUI(self, key, value, typ):
        try:
            if (typ == "BOOL"):
                value = bool(value)
            elif (typ == "INT"):
                value = int(value)
            elif (typ == "FLOAT"):
                value = float(value)
        except:
            self.logger.warning("could not cast value %s to type %s"%(str(value), str(typ)))
            self.syncModel()
            return
        self.params[key] = value

    def syncModel(self):
        self.clearModel()        
        for key, value in self.params.getParams().items():
            item = {}
            item[ParamsQML.IDX_KEY] = key
            item[ParamsQML.IDX_VALUE] = value
            if type(value) == int:
                item[ParamsQML.IDX_TYPE] = "INT"
            elif type(value) == float:
                item[ParamsQML.IDX_TYPE] = "FLOAT"
            elif type(value) == str:
                item[ParamsQML.IDX_TYPE] = "STRING"
            elif type(value) == bool:
                item[ParamsQML.IDX_TYPE] = "BOOL"
            else:
                item[ParamsQML.IDX_TYPE] = type(value).__name__
            self.add_item(item)
        self.sortModel()
        self.endResetModel()

    def sortModel(self):
        sortedModel = []
        types = ["BOOL", "INT", "FLOAT", "STRING"]
        for t in types:
            typModels = []
            for item in self.items:
                if item[ParamsQML.IDX_TYPE] == t:
                    typModels.append(item)
            
            if typModels == []:
                continue
            alphabetical = sorted(typModels, key=lambda k: k[ParamsQML.IDX_KEY])
            sortedModel.extend(alphabetical)
            
        self.items = sortedModel

    def add_item(self, item):
        self.items.append(item)

    def clearModel(self):
        #call self.endResetModel() after finishing with model
        self.beginResetModel()
        self.items = []

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self.items)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        try:
            rv =  self.items[index.row()][role]
            return rv
        except IndexError:
            return QtCore.QVariant()

    def roleNames(self):
        return {
            self.IDX_KEY: QtCore.QByteArray(b"key"),
            self.IDX_VALUE: QtCore.QByteArray(b"value"),
            self.IDX_TYPE: QtCore.QByteArray(b"type")
        }

class StepResultQML(QtCore.QObject):

    def __init__(self, stepResult):
        super(StepResultQML,self).__init__()
        self.stepResult = stepResult

    @QtCore.pyqtSlot(result=int)
    def getResult(self):
        return self.stepResult.result.value

    @QtCore.pyqtSlot(result=str)
    def getMessage(self):
        return self.stepResult.msg

    @QtCore.pyqtSlot(result=str)
    def getFullTCName(self):
        return self.stepResult.case.getFullName()

class TestRunnerQML(QtCore.QObject):

    stepStart = QtCore.pyqtSignal(str, int, int, name='stepStart')
    stepComplete = QtCore.pyqtSignal(TestSuite.SuiteState, int, str, str, int, int, name='stepComplete')
    sequenceComplete = QtCore.pyqtSignal(TestSuite.SuiteState, bool, float, str, name='sequenceComplete')
    sequenceStart = QtCore.pyqtSignal(TestSuite.SuiteState, name='sequenceStart')
    deviceIDSet = QtCore.pyqtSignal(str, name='deviceIDSet')
    errorOccured = QtCore.pyqtSignal(str, name='errorOccured')

    def __init__(self, testSuiteRunner, engine):
        super(TestRunnerQML,self).__init__()
        self.engine = engine
        self.testSuiteRunner = testSuiteRunner

        self.testSuiteRunner.sequenceStart.connect(self._onSequenceStartAsync)
        self.testSuiteRunner.sequenceComplete.connect(self._onSequenceCompleteAsync)
        self.testSuiteRunner.stepStart.connect(self._onStepStartAsync)
        self.testSuiteRunner.stepComplete.connect(self._onStepCompleteAsync)
        self.testSuiteRunner.deviceIDSet.connect(self._onDeviceIDSetAsync)
        self.testSuiteRunner.signalError.connect(self._onTestSuiteErrorAsync)
        self._stepResults = []
        self.stepTypes = TestSuite.StepType.LOAD
        self.suiteState = TestSuite.SuiteState.LOADING
        
    @QtCore.pyqtSlot(result=QtCore.QObject)
    def getTempParams(self):
        self.tempParams = ParamsQML(self.testSuiteRunner.testSuite.tempParams)
        return self.tempParams
    
    @QtCore.pyqtSlot()
    def abort(self):
        self.testSuiteRunner.abort()
        
    @QtCore.pyqtSlot()
    def enaCliRunner(self):
        cli = TestRunnerCli.TestSuiteCliRunner(testSuiteRunner=self.testSuiteRunner, playSound=False)
        
    @QtCore.pyqtSlot(str)
    def setStepType(self, stepTypes):
        if stepTypes.lower() == "regular":
            self.stepTypes = TestSuite.StepType.REGULAR
            self.suiteState = TestSuite.SuiteState.REGULAR
        elif stepTypes.lower() == "load":
            self.stepTypes = TestSuite.StepType.LOAD
            self.suiteState = TestSuite.SuiteState.LOADING
        elif stepTypes.lower() == "unload":
            self.stepTypes = TestSuite.StepType.UNLOAD
            self.suiteState = TestSuite.SuiteState.UNLOADING
        else:
            raise RuntimeError("unsupported stepTypes %s"%str(stepTypes))

    @QtCore.pyqtSlot(result=bool)
    def isRunning(self):
        return self.testSuiteRunner.isExecuting

    @QtCore.pyqtSlot(result=list)
    def getTestCases(self):    
        nameOfCases = []
        cases = self.testSuiteRunner.testSuite.getTCs(self.stepTypes)
        for case in cases:
            nameOfCases.append(case.doc)
        return nameOfCases

    @QtCore.pyqtSlot()
    def loadSuite(self):
        future = self.testSuiteRunner.loadSuite()
        if self.stepTypes == TestSuite.StepType.REGULAR:
            future.result(timeout=10)
            self.testSuiteRunner.run()
    
    @QtCore.pyqtSlot(result=bool)        
    def run(self):
        rv = self.testSuiteRunner.run()
        if rv == None:
            rv = False
        else:
            rv = True
        return rv
    
    @QtCore.pyqtSlot()        
    def unloadSuite(self):
        self.testSuiteRunner.unloadSuite()
        
    def _onDeviceIDSetAsync(self, deviceID):
        self.deviceIDSet.emit(deviceID)

    def _onStepStartAsync(self, state, testCase, idx, numCases):
        if state == self.suiteState:
            self.stepStart.emit(testCase.doc, idx, numCases)

    def _onIOMessageAsync(self, msg):
        self.ioMessage.emit(msg)

    def _onSequenceCompleteAsync(self, state, suiteResult):
        if state == self.suiteState:
            success = suiteResult.isSuccessful()
            self.sequenceComplete.emit(state, success, suiteResult.executionTime, str(suiteResult.suite.values))

    def _onSequenceStartAsync(self, state):
        if state == self.suiteState:
            self.sequenceStart.emit(state)

    def _onStepCompleteAsync(self, state, stepRes, idx, numCases):
        if state == self.suiteState:
            self.stepComplete.emit(state, stepRes.result.value, stepRes.msg, stepRes.case.getFullName(), idx, numCases)
            
    def _onTestSuiteErrorAsync(self, msg):
        self.errorOccured.emit(str(msg))

class OpenProductionQML(QtCore.QObject):
    def __init__(self, engine, workspace, stationName):
        super(OpenProductionQML,self).__init__()
        self.workspace = workspace
        self.stationName = stationName

    @QtCore.pyqtSlot(result=str)
    def getVersion(self):
        return misc.getVersion()

    @QtCore.pyqtSlot(result=str)
    def getWorkspace(self):
        return self.workspace

    @QtCore.pyqtSlot(result=str)
    def getStationName(self):
        return self.stationName

class StationQML(QtCore.QObject):

    productSaveComplete = QtCore.pyqtSignal(str, bool, int, name='productSaveComplete')
    asyncCallCompleted = QtCore.pyqtSignal(bool, QtCore.pyqtBoundSignal, name='asyncCallCompleted')
    productStationLoadCompleted = QtCore.pyqtSignal(name='productStationLoadCompleted')

    def __init__(self, engine):
        super(StationQML,self).__init__()
        self.engine = engine
        self.productVersions = None
        self.threadPool = ThreadPoolExecutor(5)
        self.revisionModel = None
        self.asyncCallCompleted.connect(self.onAsyncCallCompleted)
        self.ioHandler = None

    def getStation(self):
        return Station.Station.getInstance()

    def _productSaveAsync(self, productName):
        ok, rev = self.productSaveChanges(productName)
        self.productSaveComplete.emit(productName, ok, rev)
        
    @QtCore.pyqtSlot(result=QtCore.QObject)
    def getLoadedProduct(self):
        station = self.getStation()
        product = None
        if station != None:
            product = station.loadedProduct
        if product != None:
            self.testRunner = TestRunnerQML(product.productRunner.suiteRunner ,self.engine)
            self.engine.setObjectOwnership(self.testRunner, QtQml.QQmlEngine.CppOwnership)
            return self.testRunner
        return None

    @QtCore.pyqtSlot(str, str, result=QtCore.QObject)
    def loadStationAsync(self, workspace, stationName):
        self.ioHandler = IOHandlerQML()
        self.engine.setObjectOwnership(self.ioHandler, QtQml.QQmlEngine.CppOwnership)
        self.ioHandler.queryYesNoSignal.connect(self._onQueryYesNo)
        testRunner = Station.Station.loadStationAsync(workspace, stationName, ioHandler=self.ioHandler)
        self.testRunner = TestRunnerQML(testRunner ,self.engine)
        self.engine.setObjectOwnership(self.testRunner, QtQml.QQmlEngine.CppOwnership)
        return self.testRunner
    
    @QtCore.pyqtSlot(result=QtCore.QObject)
    def getIOHandler(self):
        return self.ioHandler
    
    def _onQueryYesNo(self, title, msg):
        mainWindow = self.engine.rootObjects()[0]
        ioHandlerPopUp = mainWindow.findChild(QtCore.QObject, "ioHandlerPopUp")
        ioHandlerPopUp.open()
        ioHandlerPopUp.setProperty("title", title)
        ioHandlerPopUp.setProperty("supportText", msg)
        ioHandlerPopUp.okButtonPressed.connect(self._onQueryAnswerOk)
        ioHandlerPopUp.cancelButtonPressed.connect(self._onQueryAnswerCancel)
        overlay = mainWindow.findChild(QtCore.QObject, "disableOverlay")
        overlay.setProperty("visible", True)
        
    @QtCore.pyqtSlot()
    def _onQueryAnswerOk(self):
        self.ioHandler.answerYesNo(TestSuiteIO.IOAnswer.YES)
        mainWindow = self.engine.rootObjects()[0]
        overlay = mainWindow.findChild(QtCore.QObject, "disableOverlay")
        overlay.setProperty("visible", False)
        
    @QtCore.pyqtSlot()
    def _onQueryAnswerCancel(self):
        self.ioHandler.answerYesNo(TestSuiteIO.IOAnswer.NO)
        mainWindow = self.engine.rootObjects()[0]
        overlay = mainWindow.findChild(QtCore.QObject, "disableOverlay")
        overlay.setProperty("visible", False)        
    
    def onAsyncCallCompleted(self, success, successSig):
        if success:
            self.popup.hide()
            successSig.emit()
        else:
            self.popup.error()
            
    
    def futureComplete(self, errValidateCB, successSig, future):
        self.asyncCallCompleted.emit(not errValidateCB(future), successSig)
        
    
    @QtCore.pyqtSlot(str, str, int)
    def loadProduct(self, productName, productVersion, revision):
        self.popup = BusyPopUp(self.engine)
        self.popup.setCaption("Lade %s '%s' (Rev %d)"%(productName, productVersion, revision))
        fut = self.threadPool.submit(self.getStation().loadProduct, productName, productVersion, revision)
        
        def errValidate(future):
            if future.exception() == None:
                if future.result() == None:
                    return True
                return False
            else:
                return True
        
#        fut.add_done_callback(lambda future: self.asyncCallCompleted.emit(future, errValidate, self.productStationLoadCompleted))
        fut.add_done_callback(functools.partial(self.futureComplete, errValidate, self.productStationLoadCompleted))

    @QtCore.pyqtSlot(result=list)
    def getProducts(self):
        rv, products = self.getStation().listProducts()
        if products == None:
            products = []
        return products

    @QtCore.pyqtSlot(result=QtCore.QObject)
    def getProductModel(self):
        self.productModel = ProductModel()
        self.engine.setObjectOwnership(self.productModel, QtQml.QQmlEngine.CppOwnership)

        rv, products = self.getStation().listProducts()
        
        if products != None:
            for prod in products:
                item = {}
                item[ProductModel.IDX_NAME] = prod["name"]
                item[ProductModel.IDX_DETAILS] = prod["description"]
                self.productModel.add_item(item)

        return self.productModel

    @QtCore.pyqtSlot(str, result=QtCore.QObject)
    def getVersionModel(self, productName):
        self.versionModel = VersionModel()
        self.engine.setObjectOwnership(self.versionModel, QtQml.QQmlEngine.CppOwnership)

        rv, self.productVersions = self.getStation().listProductVersions(productName)
        
        if self.productVersions != None:
            for vers in self.productVersions:
                item = {}
                item[VersionModel.IDX_NAME] = vers["version"]
                item[VersionModel.IDX_COMMENT] = vers["description"]
                base64Image = base64.b64encode(vers["image"])
                item[VersionModel.IDX_IMAGE] = "data:image/png;base64,"+base64Image.decode()
                self.versionModel.add_item(item)



        return self.versionModel

    @QtCore.pyqtSlot(result=QtCore.QObject)
    def getRevisionModel(self):
        if self.revisionModel != None:
            del self.revisionModel
        self.revisionModel = RevisionModel()
        self.engine.setObjectOwnership(self.revisionModel, QtQml.QQmlEngine.CppOwnership)
        return self.revisionModel



    @QtCore.pyqtSlot(str, str)
    def reloadRevisionModel(self, productName, productVersion):
        if self.revisionModel != None:
            self.revisionModel.clearModel()

        rv, revisions = self.getStation().listProductRevisions(productName, productVersion)
        if revisions != None:
            for rev in revisions:
                item = {}
                item[RevisionModel.IDX_NAME] = rev["revision_id"]
                item[RevisionModel.IDX_DATE] = rev["date"].strftime('%Y-%m-%d %H:%M')
                self.revisionModel.add_item(item)

        self.revisionModel.endResetModel()
        
class BusyPopUp(QtCore.QObject):

    def __init__(self, engine):
        super(BusyPopUp,self).__init__()
        self.engine = engine
        
        self.mainWindow = engine.rootObjects()[0]
        self.busyWindow = engine.rootObjects()[1]
        self.busyWindow.setProperty("visible", True)
        
        
        widthMain = self.mainWindow.property("width")
        width = self.busyWindow.property("width")
        xMain = self.mainWindow.property("x")
        x = round((widthMain - width) / 2)
        
        heightMain = self.mainWindow.property("height")
        height = self.busyWindow.property("height")
        yMain = self.mainWindow.property("y")
        y = round((heightMain - height) / 2)        
        
        self.busyWindow.setProperty("x", x+xMain)
        self.busyWindow.setProperty("y", y+yMain)
        self.busyWindow.errorBtnPressed.connect(self.onErrorButtonPressed)
        
        self.overlay = self.mainWindow.findChild(QtCore.QObject, "disableOverlay")
        self.overlay.setProperty("visible", True)
        
    @QtCore.pyqtSlot()
    def onErrorButtonPressed(self):
        self.mainWindow.onErrorOccured()
        self.hide()
        
    def setCaption(self, caption):
        self.busyWindow.setProperty("caption", caption)

    def error(self):
        self.busyWindow.setErrorState()
        
    def hide(self):
        self.busyWindow.setProperty("visible", False)
        self.overlay.setProperty("visible", False)
        
        
class PlaySound(QtCore.QObject):
    def __init__(self, rootPath):
        super(PlaySound,self).__init__()
        self.folder = os.path.join(rootPath, "sound")
    
    @QtCore.pyqtSlot()    
    def playSuccess(self):
        QtMultimedia.QSound.play(os.path.join(self.folder, "success.wav"))

    @QtCore.pyqtSlot()
    def playError(self):
        QtMultimedia.QSound.play(os.path.join(self.folder, "error.wav"))
        
    @QtCore.pyqtSlot()
    def playNotify(self):
        QtMultimedia.QSound.play(os.path.join(self.folder, "notify.wav"))
        
class MyImageProvider:
    
    instance = None
    
    def __init__(self):
        if not MyImageProvider.instance:
            MyImageProvider.instance = MyImageProvider.__MyImageProvider()
            
    def __getattr__(self, name):
        return getattr(self.instance, name)
    def __setattr__(self, name, value):
        return setattr(self.instance, name, value)
    
    class __MyImageProvider(QtQuick.QQuickImageProvider):
        
        def __init__(self):
            super(MyImageProvider.__MyImageProvider, self).__init__(QtQuick.QQuickImageProvider.Image) 
            self.imgs = queue.Queue()
            
        def pushImage(self, qImage):
            self.imgs.put(qImage)
    
        def requestImage(self, p_str, size):
            image = self.imgs.get()
            #keep QImage object in python scope
            self._outimg = image
            return self._outimg, self._outimg.size()

    @staticmethod
    def getInstance():
        return MyImageProvider.instance
    
        
class UIRunner:
    def __init__(self, workspace):
        log.initLogger()
        
        self.logger = logging.getLogger(misc.getAppName())
        
        self.logger.info("trying to open workspace config file")
        f = open(os.path.join(workspace, "config.json"))
        workspaceCfg = json.load(f)
        f.close()
        
        stationName = workspaceCfg["station"]
        self.logger.info("station name is %s"%stationName)

        myPath = os.path.split(os.path.abspath(__file__))[0]
        qmlFolder = os.path.join(myPath, "qml")

        self.qmlEngine = QtQml.QQmlApplicationEngine()
        self.qmlStation = StationQML(self.qmlEngine)

        context = self.qmlEngine.rootContext()
        context.setContextProperty("station", self.qmlStation)

        self.openProduction = OpenProductionQML(self.qmlEngine, workspace, stationName)
        context.setContextProperty("openProduction", self.openProduction)

        self.helper = Helper()
        context.setContextProperty("helper", self.helper)


        logger = logging.getLogger(misc.getAppName())
        self.logStream = MyLogger()
        self.logStream.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
        logger.addHandler(self.logStream)
        context.setContextProperty("logger", self.logStream)


        self.playSound = PlaySound(qmlFolder)
        context.setContextProperty("playSound", self.playSound)
        
        self.imgProvider = MyImageProvider()
        self.qmlEngine.addImageProvider("myProvider", self.imgProvider.instance)

        # Load the qml file into the engine
        f = os.path.join(qmlFolder, "main.qml")
        self.qmlEngine.load(f)

        qmlFolder = os.path.join(qmlFolder, "Common")
        f = os.path.join(qmlFolder, "BusyIndicationWindow.qml")
        self.qmlEngine.load(f)




import os
import json
from openProduction.connectors import ConnectionManager
from openProduction.station import Station
from openProduction.product import ProductExecutor
from openProduction.connectors import BaseConnector
from openProduction.server import ServerInterface
from openProduction.common import misc
import logging

class StationLoader(ProductExecutor.ProductExecutor):
    """Station Loader class"""
    
    def __init__(self, ctor):
        super(StationLoader, self).__init__(ctor)
        self.logger = logging.getLogger(misc.getAppName())
        self.workspace = self.params["workspace"]
        self.name = self.params["stationName"]
        self.cfgFile = os.path.join(self.workspace, "config.json")
            
    def step__00100(self, params, values):
        """load config file"""
        self.config = None
        
        self.ioHandler.message("trying to open file %s"%os.path.abspath(self.cfgFile))

        f = open(self.cfgFile)
        
        self.logger.info("trying to parse content of config file to json")
        self.config = json.load(f)
        self.logger.info("checking config file for required keywords")
        rv = self._checkCfg(self.config)
        self.assertEqual(rv, True, "station configuration file is invalid")

    def step__00200(self, params, values):
        """load connector"""
        connectorName = self.config["connectors"][0]["connectorName"]
        connectorCfg = self.config["connectors"][0]["connectorConfig"]
        
        self.logger.info("trying to load connector '%s'"%connectorName)

        c = ConnectionManager.ConnectionManager()
        rv, self.connector = c.loadConnector(connectorName, connectorCfg, self.workspace)
        
        if rv == BaseConnector.ConnectorErrors.INVALID_CONFIG:
            self.assertNotEqual(rv, BaseConnector.ConnectorErrors.INVALID_CONFIG, "connector config invalid: %s"%self.connector.config.errmsg())
        
        self.assertNotEqual(rv, BaseConnector.ConnectorErrors.NOT_FOUND.value, "database server not found, check IP address or hostname of server")
        self.assertNotEqual(rv, BaseConnector.ConnectorErrors.FORBIDDEN, "Connection with sql server failed, check username/password")
        
        self.assertEqual(rv, BaseConnector.ConnectorErrors.NO_ERROR, "loading connector '%s' failed with msgcode %d (%s)"%(connectorName, rv.value, rv.name))
        self.assertNotEqual(self.connector, None)
        
    def step__00300(self, params, values):
        """check if station exists"""
        s = ServerInterface.ServerInterface.create(self.workspace)
        rv, data = s.listStations()
        self.assertEqual(rv, BaseConnector.ConnectorErrors.NO_ERROR, "query station names failed with %s"%rv.name)
        names = [x["name"] for x in data]
        nameFound = self.name in names
        self.assertEqual(nameFound, True, "station %s not found in database, available stations are %s"%(self.name, names))
        for d in data:
            if d["name"] == self.name:
                self.stationID = d["station_id"]
        
    def step__00400(self, params, values):
        """construct station"""
        self.logger.info("trying to construct station '%s'"%self.name)        
        self.station = Station.Station(self.workspace, self.name, self.config, self.connector, self.stationID, ioHandler=self.ioHandler)
        self.assertNotEqual(self.station, None, "constructing station %s failed"%self.name)
        
        
    def _checkCfg(self, config):
        ok = True
        #TODO: use json schme validation
#        for keyword in self.REQUIRED:
#            if keyword not in config:
#                self.logger.error("keyword '%s' required but not found in station config %s"%(keyword, self.cfgFile))
#                ok = False
        return ok
                


import json
import os
import logging
from openProduction.common import misc, Version
from openProduction.product import Product
import traceback
from openProduction.station import StationLoader
from openProduction.suite import TestSuite, TestLoader, TestParams
from openProduction.product import ProductExecutorCtor
from openProduction.server import ServerInterface
from openProduction.connectors import BaseConnector

class Station:
    
    instance = None
    
    def __init__(self, workspace, stationName, config, resultConnector, stationID, ioHandler=None):
        if not Station.instance:
            Station.instance = Station.__Station(workspace, stationName, config, resultConnector, stationID, ioHandler=ioHandler)
            
    def __getattr__(self, name):
        return getattr(self.instance, name)
    def __setattr__(self, name, value):
        return setattr(self.instance, name, value)
    
    class __Station:
        def __init__(self, workspace, stationName, config, resultConnector, stationID, ioHandler=None):            
            self.logger = logging.getLogger(misc.getAppName())
            self.workspace = workspace
            self.stationName = stationName
            self.config = config
            self.resultConnector = resultConnector
            self.loadedProduct = None
            self.stationID = stationID
            self.serverIF = ServerInterface.ServerInterface.create(workspace)
            self.productData = None
            self.ioHandler = ioHandler
            rv, data = self.serverIF.listStationProducts(self.stationID)
            if rv == BaseConnector.ConnectorErrors.NO_ERROR:
                self.productData = data
            
        def getStationName(self):
            return self.stationName
        
        def getProductData(self):
            if self.productData == None:
                rv, data = self.serverIF.listStationProducts(self.stationID)
                if rv == BaseConnector.ConnectorErrors.NO_ERROR:
                    self.productData = data
            else:
                rv = BaseConnector.ConnectorErrors.NO_ERROR
                
            return rv, self.productData
    
        def listProducts(self):
            return self.getProductData()
        
        def getProductID(self, productName):
            productID = None
            rv, productData = self.getProductData()
            if productData != None:
                for p in productData:
                    if p["name"] == productName:
                        productID = p["product_id"]
            return productID

        def loadProduct(self, name, version, revision=None, ui=None, discardChanges=False):
            self.loadedProduct = Product.Product.load(self, name, version, revision, ioHandler=self.ioHandler, ui=ui)
            return self.loadedProduct

        def listProductVersions(self, productName):
            productID = self.getProductID(productName)
            if productID != None:
                return self.serverIF.listStationProductSteps(self.stationID, productID)
            else:
                return BaseConnector.ConnectorErrors.NOT_FOUND, None
            
        def listProductRevisions(self, productName, versionName):
            productID = self.getProductID(productName)
            if productID != None:
                filt = "`product_id`=%d and `version` like '%s' and `station_id`=%d"%(productID, versionName, self.stationID)
                rv, data = self.serverIF.listProductSteps(filt)
                if rv == BaseConnector.ConnectorErrors.NO_ERROR:
                    productStepID = data[0]["product_step_id"]
                    filt = "`product_step_id`=%d"%productStepID
                return self.serverIF.listProductRevisions(filt)
            else:
                return BaseConnector.ConnectorErrors.NOT_FOUND, None            

        def saveStation(self):
            ok = False
            fname = os.path.join(self.workspace, self.stationName, "config.json")
            try:
                f = open(fname, "w")
                f.write(json.dumps(self.config))
                f.close()
                ok = True
            except Exception as e:
                self.logger.error("saving station failed with error message %s"%str(e))
                self.logger.info("full traceback:\n%s"%traceback.format_exc())
            return ok
    
    @staticmethod
    def getInstance():
        return Station.instance
            
    @staticmethod
    def localStationExists(workspace, name):
        p = os.path.join(workspace, name)
        if os.path.exists(p):
            return True
        else:
            return False
    
    @staticmethod
    def listStations(workspace):
        names = []
        s = ServerInterface.ServerInterface.create(workspace)
        rv, data = s.listStations()
        if rv == BaseConnector.ConnectorErrors.NO_ERROR:
            names = [x["name"] for x in data]
        return names
    
    @staticmethod
    def createStation(workspace, stationName, config):
        workdir = os.path.abspath(os.path.join(workspace, stationName))
        ok, resultConnector = StationLoader.StationLoader.loadConnector(config["connectors"], workdir)
        if ok:
            s = ServerInterface.ServerInterface.create(workspace)
            rv, data = s.listStations("name like '%s'"%stationName)
            if rv == BaseConnector.ConnectorErrors.NO_ERROR:
                stationID = data[0]["station_id"]
                return Station(workspace, stationName, config, resultConnector, stationID)
        return None

    @staticmethod
    def loadStationAsync(workspace, stationName, ioHandler=None, ui=None):
        
        params = TestParams.TestParams({"workspace": workspace, 
                                        "stationName": stationName},
                                        None)

        suite = TestSuite.TestSuite("station loader",  "", params=params, values=None)
        suiteRunner = TestSuite.TestSuiteRunner(suite, stopOnFail=True, stopOnExcept=True)
        
        loader = TestLoader.TestModuleLoader(StationLoader)
        ctor = ProductExecutorCtor.ProductExecutorCtor(ioHandler, params, suite.values, suite.tempParams, ui)
        tgs = loader.discover(ctor)
        for tg in tgs:
            suite.addTestCaseGroup(tg)
        
        return suiteRunner

    @staticmethod
    def loadStation(workspace, stationName, ioHandler=None, ui=None):
        runner = Station.loadStationAsync(workspace, stationName, ioHandler, ui)
        fut = runner.loadSuite()
        fut.result(timeout = 60)
        fut = runner.run()
        fut.result(timeout = 60)
        return Station.getInstance()


from openProduction.product import ProductExecutor

class HardwareCleanUp(ProductExecutor.ProductExecutor):
    """Result connector setup hooks"""
    
    def __init__(self, ctor):
        super(HardwareCleanUp, self).__init__(ctor)

    def unload_hardware_manager_00100(self, params, values):
        """Hardwarereferenzen löschen"""
        self.hardwareManager.stop()


import logging
from openProduction.common import misc, Signals, Exceptions, Version
from openProduction.suite import TestSuiteValueContainer, TestSuiteIO, TestClass, TestParams
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
import time
import traceback


class InstallHints:
    def __init__(self, shouldBeLast=False, mustBeLast=False):
        self.mustBeLast = mustBeLast
        self.shouldBeLast = shouldBeLast

class StepType(Enum):
    SETUP = 0
    REGULAR = 1
    TEARDOWN = 2
    LOAD = 3
    UNLOAD = 4
    
class SuiteState(Enum):
    LOADING = 0
    REGULAR = 1
    UNLOADING = 2

class ExecutionResult(Enum):
    SUCCESS = 0
    FAIL = 1
    EXCEPTION = 2
    ABORT = 3
    SKIP = 4
    NOTEXECUTED = 5

class StepResult:
    def __init__(self, result, case, msg, duration):
        self.result = result
        self.case = case
        self.msg = msg
        self.duration = duration
            
    def __str__(self):
        return "result=%s, name=%s, msg=%s, duration=%.02f"%(self.result, self.case.getFullName(), self.msg, self.duration)
    
    def toDict(self):
        return {"result": self.result.value, "msg": self.msg, "duration": self.duration, "tc": self.case.toDict()}


class TestSuiteLogger(logging.Handler):

    def __init__(self):
        logging.Handler.__init__(self)
        self.txt = ""

    def emit(self, record):
        if self.txt != "":
            self.txt += "\n"
        self.txt += self.format(record)

class SuiteResult:
    def __init__(self, suite):
        self.numTotal = 0
        self.numErrors = 0
        self.numFails = 0
        self.numSkips = 0
        self.numAbort = 0
        self.numGood = 0
        self.stepResults = []
        self.executionTime = 0
        self.suite = suite
        
        logger = logging.getLogger(misc.getAppName())
        self.testSuiteLogger = TestSuiteLogger()
        self.testSuiteLogger.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
        logger.addHandler(self.testSuiteLogger)

    def addStepResult(self, stepResult):
        self.numTotal += 1
        self.stepResults.append(stepResult)
        if stepResult.result == ExecutionResult.EXCEPTION:
            self.numErrors +=1
        elif stepResult.result == ExecutionResult.FAIL:
            self.numFails += 1
        elif stepResult.result == ExecutionResult.SKIP:
            self.numSkips += 1
        elif stepResult.result == ExecutionResult.SUCCESS:
            self.numGood += 1
        elif stepResult.result == ExecutionResult.ABORT:
            self.numAbort += 1
            
    def finalize(self, executionTime):
        self.executionTime = executionTime
        logger = logging.getLogger(misc.getAppName())
        logger.removeHandler(self.testSuiteLogger)
    
    def isSuccessful(self):
        if self.numErrors != 0 or self.numFails != 0 or self.numAbort != 0:
            return False
        else:
            return True
        
    def toDict(self):
        rv = {"numTotal": self.numTotal, "numErrors": self.numErrors, "numFails": self.numFails, 
              "numSkips": self.numSkips, "numGood": self.numGood, "numAbort": self.numAbort,
              "executionTime": self.executionTime, "stepResults": []}
        for res in self.stepResults:
            rv["stepResults"].append(res.toDict())
        return rv
    
    def getLogText(self):
        return self.testSuiteLogger.txt

class TestCase:
    def __init__(self, name, doc, func, typ, installHint):
        self.name = name
        self.doc = doc
        self.func = func
        self.typ = typ
        self.group = None
        self.installHint = installHint
        
    def getFullName(self):
        name = ""
        if self.group != None:
            name += self.group.name + "."
        name += self.name
        return name
    
    def toDict(self):
        rv =  {"name": self.name, "doc": self.doc, "typ": self.typ.value}
        if self.group != None:
            rv["group"] = {"name":self.group.name, "doc":self.group.doc}
        return rv


class TestCaseSort:
    @staticmethod
    def sortByHints(testCases):
        logger = logging.getLogger(misc.getAppName())
        idx=0
        for tc in testCases:
            if tc.installHint.shouldBeLast == True:
                #push element to end
                testCases.insert(len(testCases), testCases.pop(idx))
            idx+=1
            
        idx=0
        numLast=0
        for tc in testCases:
            if tc.installHint.mustBeLast == True:
                #push element to end
                testCases.insert(len(testCases), testCases.pop(idx))
                numLast+=1
            idx+=1
            
        if numLast > 1:
            logger.warning("more than one test case marked as 'mustBeLast', ordering may be unexpected")
        
        return testCases

class TestCaseGroup:
    def __init__(self, name, doc, testClass):
        self.logger = logging.getLogger(misc.getAppName())
        if isinstance(testClass, TestClass.TestClass) == False:
            raise RuntimeError ("testClass must be of type openProduction.suite.TestClass.TestClass")
        
        self.testClass = testClass
        self.name = name
        self.doc = doc
        self._tcs = []
        self.runnerInstalled = False
    
    def install(self, suite):
        return self.testClass.install(suite)
    
    def addTestCase(self, tc, index=None):
        tc.group = self
        if index == None:
            index=len(self._tcs)
        self._tcs.insert(index, tc)
        
    def getFullName(self):
        return self.name
    
    def getNumTCs(self):
        return len(self._tcs)
    
    def getTCs(self, typ, sort=True):
        myTcl = []
        for t in self._tcs:
            if t.typ == typ:
                myTcl.append(t)
                
        if sort:
            def myFunc(obj):
                return obj.getFullName()
            myTcl.sort(key=myFunc)
            
        myTcl = TestCaseSort.sortByHints(myTcl)
                
        return myTcl

    def getTestCases(self, sort=True):
        return self.getTCs(StepType.REGULAR, sort)
        
    def getSetUps(self, sort=True):
        return self.getTCs(StepType.SETUP, sort)
                
    def getTearDowns(self, sort=True):
        return self.getTCs(StepType.TEARDOWN, sort)
    
    def getLoaders(self, sort=True):
        return self.getTCs(StepType.LOAD, sort)
    
    def getUnLoaders(self, sort=True):
        return self.getTCs(StepType.UNLOAD, sort)

class TestSuite:
    def __init__(self, name, doc, params=None, values=None, tempParams=None):
        self.logger = logging.getLogger(misc.getAppName())
        self.name = name
        self.doc = doc
                    
        self.runner = None
        
        if params != None:
            if isinstance(params, TestParams.TestParams) == False:
                raise RuntimeError("params has to be of type TestParams.TestParams")
        if values != None:
            if isinstance(values, TestParams.TestParams) == False:
                raise RuntimeError("values has to be of type TestParams.TestParams")
        if tempParams != None:
            if isinstance(tempParams, TestParams.TestParams) == False:
                raise RuntimeError("tempParams has to be of type TestParams.TestParams")

            
        self.params = params
        if values == None:
            self.values = TestParams.TestParams({}, None)
        else:
            self.values = values
        if tempParams == None:
            self.tempParams = TestParams.TestParams({}, None)
        else:
            self.tempParams = tempParams
        self._tcg = []
        
    def _validate(self):
        tcLastName = ""
        funcs = [self.getTestCases, self.getSetUps, self.getTearDowns, 
                 self.getLoaders, self.getUnLoaders]
        
        for func in funcs:
            anyML = False
            tcg = func()
            for tc in tcg:
                if tc.installHint.mustBeLast == True:
                    if anyML == True:
                        raise RuntimeError("both testcases %s and %s have installation hint 'mustBeLast'"%(tcLastName, tc.name))
                    anyML = True
                    tcLastName = tc.name
                    
        return True
    
    def abort(self):
        for tg in self._tcg:
            tg.testClass.abort()

    def abortClear(self):
        for tg in self._tcg:
            tg.testClass.abortClear()

    def unload(self):
        for tg in self._tcg:
            tg.testClass.unload()

    def unloadClear(self):
        for tg in self._tcg:
            tg.testClass.unloadClear()
        
    def installRunner(self, runner):
        self.runner = runner
        for tcg in self._tcg:
            if tcg.testClass.runnerInstalled == False:
                rv = tcg.testClass.installRunner(runner)
                if rv == False:
                    raise RuntimeError("intallation of runner@group %s failed"%tcg.name)
                tcg.testClass.runnerInstalled = True
        
    def addMethodHint(self, func, installHint):
        tc = self.getTCBYFunc(func)
        if tc != None:
            tc.installHint = installHint
            
        self._validate()
    
    def addTestCaseGroup(self, tcg, index=None):
        rv = None
        if isinstance(tcg, TestCaseGroup):
            self._tcg.append(tcg)
            rv = tcg.install(self)
            if rv != False:
                self._validate()
                if self.runner != None:
                    rv = tcg.testClass.installRunner(self.runner)
                    if rv == False:
                        raise RuntimeError("intallation of runner@group %s failed"%tcg.name)
                    tcg.testClass.runnerInstalled = True
        else:
            raise RuntimeError("param tcg has to be a TestCaseGroup")
        return rv
        
    def getTCBYFunc(self, func):
        rv = None
        for tg in self._tcg:
            for tc in tg._tcs:
                if tc.func == func:
                    rv = tc
                    break
        return rv
    
    def getNumTCs(self):
        numCases = 0
        for tg in self._tcg:
            numCases += tg.getNumTCs()    
        return numCases
        
    def getTCs(self, typ, sort=True):
        myTcl = []
        for tg in self._tcg:
            tcs = tg.getTCs(typ, sort)
            for tc in tcs:
                myTcl.append(tc)
                
        if sort:
            def myFunc(obj):
                return obj.getFullName()
            myTcl.sort(key=myFunc)
            
        myTcl = TestCaseSort.sortByHints(myTcl)
            
        return myTcl

    def getTestCases(self, sort=True):
        return self.getTCs(StepType.REGULAR, sort)
        
    def getSetUps(self, sort=True):
        return self.getTCs(StepType.SETUP, sort)
                
    def getTearDowns(self, sort=True):
        return self.getTCs(StepType.TEARDOWN, sort)
    
    def getLoaders(self, sort=True):
        return self.getTCs(StepType.LOAD, sort)
    
    def getUnLoaders(self, sort=True):
        return self.getTCs(StepType.UNLOAD, sort)
    
    def getTestCaseGroups(self):
        return self._tcg
    
class TestSuiteRunner:
    
    def __init__(self, testSuite, stopOnFail=True, stopOnExcept=True):
        self.logger = logging.getLogger(misc.getAppName())
        
        self.testSuite = testSuite
        
        self.sequenceStart = Signals.Signal()
        self.sequenceComplete = Signals.Signal()
        self.stepStart = Signals.Signal()
        self.stepComplete = Signals.Signal()
        self.signalError = Signals.Signal()
        self.prepareNextRun = Signals.Signal()
        self.keyChanged = Signals.Signal()
        self.deviceIDSet = Signals.Signal()
        
        testSuite.values.keyChanged.connect(self.keyChanged)
        testSuite.values.deviceIDSet.connect(self.deviceIDSet)
        
        self._stopOnFail = stopOnFail
        self._stopOnExcept = stopOnExcept
        self.isExecuting = False
        self.isLoaded = False
        
        self._pool = ThreadPoolExecutor(3)
        
        self.testSuite.installRunner(self)
        
        if self.testSuite.params != None:
            self.testSuite.params.setReadOnly(True)
        
        self.suiteResults = None
        self._abortRun = False
            
    def _executeCases(self, executionType, cases, stopOnExcept=True, stopOnFail=True):
        rv = True
        
        curIdx = 0
        for case in cases:
            self.stepStart.emit(executionType, case, curIdx, len(cases))
            stepStart = time.time()
                
            try:
                res = ExecutionResult.SUCCESS
                msg = "Ok"
                rv = case.func(self.testSuite.params, self.testSuite.values)
                if rv != None:
                    if rv == ExecutionResult.SKIP:
                        res = ExecutionResult.SKIP
                        msg = "Skip"
                    elif isinstance(rv, TestCaseGroup):
                        self.testSuite.addTestCaseGroup(rv)
                        self.logger.info("test case group %s added to test suite"%rv.getFullName())
                    elif isinstance(rv, TestSuiteIO.IOMessage):
                        msg = rv.msg
                    elif type(rv) == type([]):
                        for val in rv:
                            if isinstance(val, TestCaseGroup):
                                self.testSuite.addTestCaseGroup(val)
                                self.logger.info("test case group %s added to test suite"%val.getFullName())
                    else:
                        pass
                rv = True
                if case.installHint.mustBeLast == True and curIdx+1 != len(cases):
                    msg = "case.mustBeLast set, but idx=%d of %d"%(curIdx+1, len(cases))
                    res = ExecutionResult.FAIL
                    rv = False
                
            except Exceptions.OpenProductionAssertionError as e:
                msg = str(e)
                res = ExecutionResult.FAIL
                rv = False
            except Exceptions.OpenProductionAbortError:
                msg = traceback.format_exc()
                res = ExecutionResult.ABORT
                rv = False
            except:
                msg = traceback.format_exc()
                res = ExecutionResult.EXCEPTION
                rv = False
                
            if (res == ExecutionResult.SUCCESS or res == ExecutionResult.SKIP) and self._abortRun == True:
                res = ExecutionResult.ABORT
                msg = "Abort"
                
            stepEnd = time.time()
            duration = stepEnd - stepStart
            
            result = StepResult(res, case, msg, duration)
            self.suiteResults.addStepResult(result)
                
            self.stepComplete.emit(executionType, result, curIdx, len(cases))
            
            curIdx+=1
            
            if res == ExecutionResult.EXCEPTION and stopOnExcept:
                break
            elif res == ExecutionResult.FAIL and stopOnFail:
                break
            elif res == ExecutionResult.ABORT:
                break
            
        return rv
        
    def _executeSuite(self, executionType):      
        executionStart = time.time()
        self.suiteResults = SuiteResult(self.testSuite)
        self.testSuite.values.clear()
        
        skipTearDown = False
        self.sequenceStart.emit(executionType)
        
        resultsFinalized = False
        
        if executionType == SuiteState.REGULAR:
            cases = self.testSuite.getSetUps()
            self._executeCases(executionType, cases, stopOnExcept=True, stopOnFail=True)
            
            if self.suiteResults.isSuccessful():
                cases = self.testSuite.getTestCases()
                self._executeCases(executionType, cases, self._stopOnExcept, self._stopOnFail)
            else:
                skipTearDown = True
                
            resultsFinalized = True
            executionEnd = time.time()
            duration = executionEnd - executionStart
            self.suiteResults.finalize(duration)
                        
            if skipTearDown == False:
                cases = self.testSuite.getTearDowns()
                self._executeCases(executionType, cases, stopOnExcept=False, stopOnFail=False)
        elif executionType == SuiteState.LOADING:
            cases = self.testSuite.getLoaders()
            self._executeCases(executionType, cases, stopOnExcept=True, stopOnFail=True)
        elif executionType == SuiteState.UNLOADING:
            cases = self.testSuite.getUnLoaders()
            self._executeCases(executionType, cases, stopOnExcept=False, stopOnFail=False)
        
        
        if resultsFinalized == False:
            executionEnd = time.time()
            duration = executionEnd - executionStart
            self.suiteResults.finalize(duration)
                
        if executionType == SuiteState.LOADING and self.suiteResults.isSuccessful():
            self.isLoaded = True
        elif executionType == SuiteState.UNLOADING:
            self.isLoaded = False

        self.testSuite.abortClear()
        self._abortRun = False
        self.isExecuting = False
        
        self.sequenceComplete.emit(executionType, self.suiteResults)
        
        if (executionType == SuiteState.REGULAR):
            self.prepareNextRun.emit(self.suiteResults)
        
        return self.suiteResults
    
    def abort(self):
        self._abortRun = True
        self.testSuite.abort()
    
    def loadSuite(self):
        return self._execute(SuiteState.LOADING, msg="Can't load suite, the product is still running")
        
    def unloadSuite(self):
        return self._execute(SuiteState.UNLOADING, msg="Can't unload suite, the product is still running")

    def run(self):
        if self.isLoaded == False:
            self.logger.warning("Can't run an unloaded test suite")
            return None
        return self._execute(SuiteState.REGULAR)
        
    def _execute(self, executionType, msg="Can't start product run, the product is still running"):
        #run test suite methods asynchronously
        if self.isExecuting == True:
            self.logger.warning(msg)
            return None
        
        if executionType == SuiteState.LOADING:
            self._abortRun = False
            self.testSuite.abortClear()
            self.testSuite.unloadClear()
        elif executionType == SuiteState.UNLOADING:
            self.testSuite.unload()
        
        self.isExecuting = True
        future = self._pool.submit(self._executeSuite, (executionType))
        future.add_done_callback(self._onExecutionDone)
        return future

    def _onExecutionDone(self, fut):
        ex = fut.exception()
        
        if ex != None:
            self.logger.error(ex)
            self.signalError.emit(ex)


if __name__ == "__main__":
    pass


import os
import traceback
from openProduction.common import Version, Constants, misc, SCM
from openProduction.product import ProductExecutor
from openProduction.suite import TestRunnerCli, TestParams, TestSuiteIO
import logging
import distutils.dir_util
from openProduction.connectors.BaseConnector import ConnectorErrors

class Product:
    def __init__(self, productData, revisionData, folder, ioHandler=None, ui=None):
        self.logger = logging.getLogger(misc.getAppName())
        
        version = productData["version"]
        revision = revisionData["revision_id"]
        url = productData["git_url"]
        
        self.version = Version.Version(version, revision, url)
        self.productID = productData['product_id']
        self.productStepID = productData['product_step_id']
            
        self.folder = folder
            
        self.discover(folder, revisionData["params"])
            
        self.name = productData["name"]
        
        if ioHandler == None:
            ioHandler = TestSuiteIO.BaseIOHandler()

        self._iohandler = ioHandler
        self._ui = ui
        
        self._createProductRunner(folder)
        
    def discover(self, folder, params):
        self.params = TestParams.TestParams(params, folder)
  
    @staticmethod
    def load(station, name, version, revision, ioHandler=None, ui=None):
        ok = True
        p = None
        logger = logging.getLogger(misc.getAppName())
        
        logger.info("getProductStepByName")
        rv, stepData = station.serverIF.getProductStepByName(station.stationName, name, version)
        if rv == ConnectorErrors.NO_ERROR:
            if revision == None:
                logger.info("getLatestProductRevision")
                rv, revData = station.serverIF.getLatestProductRevision(stepData["product_step_id"])
            else:
                logger.info("getProductRevision")
                rv, revData = station.serverIF.getProductRevision(revision)

            if rv == ConnectorErrors.NO_ERROR:    
                scmDir = os.path.join(station.workspace, "repo")
                stepData["commit_id"] = revData["commit_id"]
                
                try:
                    scm = SCM.SCM(scmDir, stepData)
                except:
                    logger.error("error creating SCM object")
                    logger.info("full traceback:\n%s"%traceback.format_exc())
                    ok = False
                
                if ok == True:
                    try:
                        scm.checkoutProductRevision()
                        ok = True
                    except:
                        logger.error("error checkout product revision")
                        logger.info("full traceback:\n%s"%traceback.format_exc())
                        ok = False
                    
                if ok == True:
                    try:
                        p = Product(stepData, revData, scmDir, ioHandler=ioHandler, ui=ui)
                    except:
                        logger.error("error loading product %s with version %s"%(name, version))
                        logger.info("full traceback:\n%s"%traceback.format_exc())
            else:
                logger.error("get latest revision for product_step_id %d failed with %s"(stepData["product_step_id"], rv.name))
        else:
            logger.error("query for product %s, version %s @ station %s failed with %s"(name, version, station.stationName, rv.name))
        return p       
    
    def _createProductRunner(self, folder):                
        self.productRunner = ProductExecutor.ProductRunner(folder, self.name, self.params,
                                                           ioHandler=self._iohandler, ui=self._ui)



from openProduction.common import Types

class Config:
    def __init__(self, description):
        #description must be a list of Types.BaseType
        if type(description) != type([]):
            raise RuntimeError("description must by a list")
            
        for e in description:
            if not isinstance(e, Types.BaseType):
                raise RuntimeError("each entry of description must be a subclass of Types.BaseType")
        
        self.__fields__ = description
        self._curIdx = 0
        self.isZombie = False
        self.err = ""
           
    def __next__(self):
        if self._curIdx >= len(self.__fields__):
            raise StopIteration()
        else:
            rv = self.__fields__[self._curIdx]
            self._curIdx += 1
            return rv
    
    def __iter__(self):
        return self
    
    def __len__(self):
        return len(self.__fields__)
    
    def fromDict(self, myDict):
        ok = True
        fieldNames = []
        for field in self.__fields__:
            fieldNames.append(field.name)
        
        dictKeys = []
        for key, val in myDict.items():
            dictKeys.append(key)
           
        missingKeys = set(fieldNames) - set(dictKeys)
             
        if len(missingKeys) > 0:
            self.err = "missing keys: "+str(missingKeys)
            ok = False
            
        if ok:
            for field in self.__fields__:
                field.value = myDict[field.name]
            
        return ok
        
    def __getitem__(self, key):
        for field in self.__fields__:
            if field.name == key:
                return field.value
    
    def errmsg(self):
        return self.err


from enum import Enum
from openProduction.common import Signals

class ProgressResult(Enum):
    INIT = 0
    SUCCESS = 1
    ERROR = 2
    UPDATE = 3

class ProgressHandler:
           
    def __init__(self):
        self.newProgress  = Signals.Signal()
        
    def reportProgress(self, idx, numSteps, description, result):
        self.newProgress.emit(idx, numSteps, description, result)
    


from openProduction.common import Constants

class Version:
    def __init__(self, version, revision, url, isDirty=False):
        if Constants.isValidStr(version) == False:
            raise RuntimeError("version string may not contain any charachter of '%s'"%Constants.invalidChars())
        
        self.version = version
        self.revision = int(revision)
        self.isDirty = isDirty
        self.url = url
        
    @staticmethod
    def fromString(myStr):
        url = ""
        if len(myStr.split(Constants.URL_SEPARATOR)) == 2:
            url, versStr = myStr.split(Constants.URL_SEPARATOR)
        elif len(myStr.split(Constants.URL_SEPARATOR)) == 1:
            versStr = myStr
        else:
            return None
        
        if len(versStr.split(Constants.VERSION_SEPARATOR)) == 2:
            version, revStr = versStr.split(Constants.VERSION_SEPARATOR)
        else:
            return None
        
        if Constants.isValidStr(version) == False:
            return None
        
        idxDirty = revStr.find("-dirty")
        if idxDirty != -1:
            revStr = revStr[:revStr.find("-dirty")]
            isDirty = True
        else:
            isDirty = False

        try:
            rev = int(revStr)
            return Version(version, rev, url, isDirty)
        except:
            pass
            
        return None
            
    def toString(self):
        vers = str(self.version) + Constants.VERSION_SEPARATOR + str(self.revision)
        if self.isDirty:
            vers = vers + "-dirty"
        return vers
        
    def toFullString(self):
        if self.url == "" or self.url == None:
            urlStr = ""
        else:
            urlStr = str(self.url) + Constants.URL_SEPARATOR
        return  urlStr + self.toString()
    
if __name__ == "__main__":
    v0 = Version.fromString("AA.34-dirty")
    v1 = Version.fromString("AA.34")
    v2 = Version.fromString("myUrlsfdsdf///.git|AA.34-dirty")


import cmd2
from openProduction.station import StationManager, Station, StationInstaller
from openProduction.product import ProductInstaller
from openProduction.common import misc
import argparse

class StationCli(cmd2.Cmd):
    intro = '===============================================================================\n\
Welcome to the openProduction station shell.   Type help or ? to list commands.\n'
    
    def __init__(self, workspace, stationName="", productName="", productVersion="", productRevision=None):
        cmd2.Cmd.__init__(self, use_ipython=True)
        self._load(workspace)
        self._loadStation(stationName)
        self._loadProduct(productName, productVersion, productRevision)
        
    def _load(self, workspace):
        self.workspace = workspace
        self.stationManager = StationManager.StationManager(self.workspace)
        self.station = None
        self.productionRun = None
        self.prompt = "> "
    
    def _loadStation(self, stationName):
        if stationName != "":
            self.station = Station.Station.loadStation(self.workspace, stationName)
            if self.station != None:
                self.prompt = "(%s) "%stationName

    def _loadProduct(self, productName, version, revision):
        if productName != "" and version != "" and self.station != None:
            self.product = self.station.loadProduct(productName, version, revision)
            if self.product != None:
                self.prompt = "(%s - %s[%s.%d]) "%(self.station.getStationName(), self.product.getProductName(), self.product.getProductVersion(), self.product.getProductRevision())

    productCreate_parser = argparse.ArgumentParser()
    productCreate_parser.add_argument('productName', type=str, help='product name')
    productCreate_parser.add_argument('productVersion', type=str, help='product version')
    @cmd2.with_argparser(productCreate_parser)
    def do_createProduct(self, arg):
        """create a new production run"""
        if self.station != None:
            installer = ProductInstaller.ProductInstaller()
            installer.run(self.station, productName=arg.productName,
                          productVersion=arg.productVersion)
        else:
            print("no station loaded")
        
    def do_getWorkspace(self, arg):
        """Get current workspace"""
        print(self.workspace)
    
    workspaceSetter_parser = argparse.ArgumentParser()
    workspaceSetter_parser.add_argument('workspace', type=str, help='workspace')
    @cmd2.with_argparser(workspaceSetter_parser)
    def do_setWorkspace(self, arg):
        """Set current workspace"""
        self._load(arg.workspace)
    
    def do_installNewStation(self, arg):
        """Install a new station"""
        installer = StationInstaller.StationInstaller()
        installer.run()
        
    def do_listStations(self, arg):
        """List all available stations"""
        stations = "\t".join(self.stationManager.listStations())
        print("Available stations on this machine: ", stations)

    loadStation_parser = argparse.ArgumentParser()
    loadStation_parser.add_argument('stationName', type=str, help='station to load')
    @cmd2.with_argparser(loadStation_parser)
    def do_loadStation(self, arg):
        """Load given station"""
        self._loadStation(arg.stationName)
            
    def do_exit(self, arg):
        'Stop recording, close the turtle window, and exit:  BYE'
        print('Thank you for using openProduction')
        return True

if __name__ == '__main__':
    pass
    #StationCli("prog").cmdloop()


class BaseType:
    def __init__(self, name, description):
        self.name = name
        if description == None:
            self.description = name
        else:
            self.description = description
        self.value = None
        
    def clear(self):
        self.value = None
        
    def __str__(self):
        return str(self.value)
            
    def setValue(self, value):
        self.value = value
        
    def getValue(self):
        return self.value
    
    def getName(self):
        return self.value

class Date(BaseType):
    def __init__(self, name, description, value=None):
        super(Date, self).__init__(name, description)
        self.value = value
        
class Path(BaseType):
    def __init__(self, name, description, value=None):
        super(Path, self).__init__(name, description)
        self.value = value
        
class String(BaseType):
    def __init__(self, name, description, value=None):
        super(String, self).__init__(name, description)
        self.value = value
        
class Int(BaseType):
    def __init__(self, name, description=None, value=None, minVal=None, maxVal=None):
        super(Int, self).__init__(name, description)
        self.value = value
        self.minVal = minVal
        self.maxVal = maxVal

class Float(BaseType):
    def __init__(self, name, description=None, value=None, minVal=None, maxVal=None):
        super(Float, self).__init__(name, description)
        self.value = value
        self.minVal = minVal
        self.maxVal = maxVal

class Bool(BaseType):
    def __init__(self, name, description=None, value=None):
        super(Bool, self).__init__(name, description)
        self.value = value

class Password(BaseType):
    def __init__(self, name, description, value=None):
        super(Password, self).__init__(name, description)
        self.value = value
        


import logging
from openProduction.common import misc

def initLogger():
    formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s')

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    logger = logging.getLogger(misc.getAppName())
    logger.setLevel(logging.DEBUG)
    
    if len(logger.handlers) == 0:
        logger.addHandler(handler)
    return logger